# Copyright (c) 2014 Quobyte Inc.
# Copyright (c) 2013 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import errno
import fnmatch
import os
import shutil

from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import fileutils
import psutil

from cinder import compute
from cinder import coordination
from cinder import exception
from cinder.i18n import _
from cinder.image import image_utils
from cinder import interface
from cinder import utils
from cinder.volume import configuration
from cinder.volume.drivers import remotefs as remotefs_drv

VERSION = '1.1.13'

LOG = logging.getLogger(__name__)

volume_opts = [
    cfg.StrOpt('quobyte_volume_url',
               help=('Quobyte URL to the Quobyte volume using e.g. a DNS SRV'
                     ' record (preferred) or a host list (alternatively) like'
                     ' quobyte://<DIR host1>, <DIR host2>/<volume name>')),
    cfg.StrOpt('quobyte_client_cfg',
               help=('Path to a Quobyte Client configuration file.')),
    cfg.BoolOpt('quobyte_sparsed_volumes',
                default=True,
                help=('Create volumes as sparse files which take no space.'
                      ' If set to False, volume is created as regular file.')),
    cfg.BoolOpt('quobyte_qcow2_volumes',
                default=True,
                help=('Create volumes as QCOW2 files rather than raw files.')),
    cfg.StrOpt('quobyte_mount_point_base',
               default='$state_path/mnt',
               help=('Base dir containing the mount point'
                     ' for the Quobyte volume.')),
    cfg.BoolOpt('quobyte_volume_from_snapshot_cache',
                default=False,
                help=('Create a cache of volumes from merged snapshots to '
                      'speed up creation of multiple volumes from a single '
                      'snapshot.')),
    cfg.BoolOpt('quobyte_overlay_volumes',
                default=False,
                help=('Create new volumes from the volume_from_snapshot_cache'
                      ' by creating overlay files instead of full copies. This'
                      ' speeds up the creation of volumes from this cache.'
                      ' This feature requires the options'
                      ' quobyte_qcow2_volumes and'
                      ' quobyte_volume_from_snapshot_cache to be set to'
                      ' True. If one of these is set to False this option is'
                      ' ignored.'))
]

CONF = cfg.CONF
CONF.register_opts(volume_opts, group=configuration.SHARED_CONF_GROUP)


@interface.volumedriver
class QuobyteDriver(remotefs_drv.RemoteFSSnapDriverDistributed):
    """Cinder driver for Quobyte USP.

    Volumes are stored as files on the mounted Quobyte volume. The hypervisor
    will expose them as block devices.

    Unlike other similar drivers, this driver uses exactly one Quobyte volume
    because Quobyte USP is a distributed storage system. To add or remove
    capacity, administrators can add or remove storage servers to/from the
    volume.

    For different types of volumes e.g., SSD vs. rotating disks,
    use multiple backends in Cinder.

    Note: To be compliant with the inherited RemoteFSSnapDriver, Quobyte
          volumes are also referred to as shares.


    .. code-block:: none

      Version history:

        1.0   - Initial driver.
        1.1   - Adds optional insecure NAS settings
        1.1.1 - Removes getfattr calls from driver
        1.1.2 - Fixes a bug in the creation of cloned volumes
        1.1.3 - Explicitely mounts Quobyte volumes w/o xattrs
        1.1.4 - Fixes capability to configure redundancy in quobyte_volume_url
        1.1.5 - Enables extension of volumes with snapshots
        1.1.6 - Optimizes volume creation
        1.1.7 - Support fuse subtype based Quobyte mount validation
        1.1.8 - Adds optional snapshot merge caching
        1.1.9 - Support for Qemu >= 2.10.0
        1.1.10 - Adds overlay based volumes for snapshot merge caching
        1.1.11 - NAS secure ownership & permissions are now False by default
        1.1.12 - Ensure the currently configured volume url is always used
        1.1.13 - Allow creating volumes from snapshots in state 'backing-up'

    """

    driver_volume_type = 'quobyte'
    driver_prefix = 'quobyte'
    volume_backend_name = 'Quobyte'
    VERSION = VERSION

    # ThirdPartySystems wiki page
    CI_WIKI_NAME = "Quobyte_CI"

    QUOBYTE_VOLUME_SNAP_CACHE_DIR_NAME = "volume_from_snapshot_cache"

    def __init__(self, execute=processutils.execute, *args, **kwargs):
        super(QuobyteDriver, self).__init__(*args, **kwargs)
        self.configuration.append_config_values(volume_opts)

        # Used to manage snapshots which are currently attached to a VM.
        self._nova = None

    @staticmethod
    def get_driver_options():
        return volume_opts

    def _create_regular_file(self, path, size):
        """Creates a regular file of given size in GiB using fallocate."""
        self._fallocate_file(path, size)

    @coordination.synchronized('{self.driver_prefix}-{snapshot.id}')
    def _delete_snapshot(self, snapshot):
        cache_path = self._local_volume_from_snap_cache_path(snapshot)
        if os.access(cache_path, os.F_OK):
            self._remove_from_vol_cache(
                cache_path,
                ".parent-" + snapshot.id, snapshot.volume)
        super(QuobyteDriver, self)._delete_snapshot(snapshot)

    def _ensure_volume_from_snap_cache(self, mount_path):
        """This expects the Quobyte volume to be mounted & available"""
        cache_path = os.path.join(mount_path,
                                  self.QUOBYTE_VOLUME_SNAP_CACHE_DIR_NAME)
        if not os.access(cache_path, os.F_OK):
            LOG.info("Volume from snapshot cache directory does not exist, "
                     "creating the directory %(volcache)s",
                     {'volcache': cache_path})
            fileutils.ensure_tree(cache_path)
        if not (os.access(cache_path, os.R_OK)
                and os.access(cache_path, os.W_OK)
                and os.access(cache_path, os.X_OK)):
            msg = _("Insufficient permissions for Quobyte volume from "
                    "snapshot cache directory at %(cpath)s. Please update "
                    "permissions.") % {'cpath': cache_path}
            raise exception.VolumeDriverException(msg)
        LOG.debug("Quobyte volume from snapshot cache directory validated ok")

    def _fallocate_file(self, path, size):
        """Calls fallocate on the given path with the given size in GiB."""
        self._execute('fallocate', '-l', '%sGiB' % size,
                      path, run_as_root=self._execute_as_root)

    def _get_backing_chain_for_path(self, volume, path):
        raw_chain = super(QuobyteDriver, self)._get_backing_chain_for_path(
            volume, path)
        # NOTE(kaisers): if the last element resides in the cache snip it off,
        # as the RemoteFS driver cannot handle it.
        if len(raw_chain) and (self.QUOBYTE_VOLUME_SNAP_CACHE_DIR_NAME in
                               raw_chain[-1]['filename']):
            del raw_chain[-1]
        return raw_chain

    def _local_volume_from_snap_cache_path(self, snapshot):
        path_to_disk = os.path.join(
            self._local_volume_dir(snapshot.volume),
            self.QUOBYTE_VOLUME_SNAP_CACHE_DIR_NAME,
            snapshot.id)

        return path_to_disk

    def _qemu_img_info_base(self, path, volume_name, basedir,
                            force_share=True,
                            run_as_root=False):
        # NOTE(kaisers): This uses a specialized backing file template in
        # order to allow for backing files in the volume_from_snapshot_cache.
        backing_file_template = remotefs_drv.BackingFileTemplate(
            "(#basedir/[0-9a-f]+/)?("
            "#volname(.(tmp-snap-)?[0-9a-f-]+)?#valid_ext|"
            "%(cache)s/(tmp-snap-)?[0-9a-f-]+(.(child-|parent-)"
            "[0-9a-f-]+)?)$" % {
                'cache': self.QUOBYTE_VOLUME_SNAP_CACHE_DIR_NAME
            })
        return super(QuobyteDriver, self)._qemu_img_info_base(
            path, volume_name, basedir, ext_bf_template=backing_file_template,
            force_share=True)

    def _remove_from_vol_cache(self, cache_file_path, ref_suffix, volume):
        """Removes a reference and possibly volume from the volume cache

        This method removes the ref_id reference (soft link) from the cache.
        If no other references exist the cached volume itself is removed,
        too.

        :param cache_file_path file path to the volume in the cache
        :param ref_suffix The id based suffix of the cache file reference
        :param volume The volume whose share defines the cache to address
        """
        # NOTE(kaisers): As the cache_file_path may be a relative path we use
        # cache dir and file name to ensure absolute paths in all operations.
        cache_path = os.path.join(self._local_volume_dir(volume),
                                  self.QUOBYTE_VOLUME_SNAP_CACHE_DIR_NAME)
        cache_file_name = os.path.basename(cache_file_path)
        # delete the reference
        LOG.debug("Deleting cache reference %(cfp)s%(rs)s",
                  {"cfp": cache_file_path, "rs": ref_suffix})
        fileutils.delete_if_exists(os.path.join(cache_path,
                                                cache_file_name + ref_suffix))

        # If no other reference exists, remove the cache entry.
        for file in os.listdir(cache_path):
            if fnmatch.fnmatch(file, cache_file_name + ".*"):
                # found another reference file, keep cache entry
                LOG.debug("Cached volume %(file)s still has at least one "
                          "reference: %(ref)s",
                          {"file": cache_file_name, "ref": file})
                return
        # No other reference found, remove cache entry
        LOG.debug("Removing cached volume %(cvol)s as no more references for "
                  "this cached volume exist.",
                  {"cvol": os.path.join(cache_path, cache_file_name)})
        fileutils.delete_if_exists(os.path.join(cache_path, cache_file_name))

    def _strip_qb_protocol(self, url):
        # Strip quobyte:// from the URL
        protocol = self.driver_volume_type + "://"
        if url.startswith(protocol):
            return url[len(protocol):]
        return url

    def do_setup(self, context):
        """Any initialization the volume driver does while starting."""
        super(QuobyteDriver, self).do_setup(context)

        self.set_nas_security_options(is_new_cinder_install=False)
        self.shares = {}  # address : options
        self._nova = compute.API()
        self.base = self.configuration.quobyte_mount_point_base
        if self.configuration.quobyte_overlay_volumes:
            if not (self.configuration.quobyte_qcow2_volumes and
                    self.configuration.quobyte_volume_from_snapshot_cache):
                self.configuration.quobyte_overlay_volumes = False
                LOG.warning("Configuration of quobyte_qcow2_volumes and "
                            "quobyte_volume_from_snapshot_cache is "
                            "incompatible with "
                            "quobyte_overlay_volumes=True. "
                            "quobyte_overlay_volumes "
                            "setting will be ignored.")

    def check_for_setup_error(self):
        if not self.configuration.quobyte_volume_url:
            msg = (_("There's no Quobyte volume configured (%s). Example:"
                     " quobyte://<DIR host>/<volume name>") %
                   'quobyte_volume_url')
            LOG.warning(msg)
            raise exception.VolumeDriverException(msg)

        # Check if mount.quobyte is installed
        try:
            self._execute('mount.quobyte', check_exit_code=False,
                          run_as_root=False)
        except OSError as exc:
            if exc.errno == errno.ENOENT:
                raise exception.VolumeDriverException(
                    'mount.quobyte is not installed')
            else:
                raise

    def set_nas_security_options(self, is_new_cinder_install):
        self._execute_as_root = False

        LOG.debug("nas_secure_file_* settings are %(ops)s (ownership) and "
                  "%(perm)s (permissions).",
                  {'ops': self.configuration.nas_secure_file_operations,
                   'perm': self.configuration.nas_secure_file_permissions}
                  )

        if self.configuration.nas_secure_file_operations == 'auto':
            LOG.debug("Mapping 'auto' value to 'false' for"
                      " nas_secure_file_operations.")
            self.configuration.nas_secure_file_operations = 'false'

        if self.configuration.nas_secure_file_permissions == 'auto':
            LOG.debug("Mapping 'auto' value to 'false' for"
                      " nas_secure_file_permissions.")
            self.configuration.nas_secure_file_permissions = 'false'

        if self.configuration.nas_secure_file_operations == 'false':
            LOG.warning("The NAS file operations will be run as "
                        "root, allowing root level access at the storage "
                        "backend.")
            self._execute_as_root = True
        else:
            LOG.info("The NAS file operations will be run as"
                     " non privileged user in secure mode. Please"
                     " ensure your libvirtd settings have been configured"
                     " accordingly (see section 'OpenStack' in the Quobyte"
                     " Manual.")

        if self.configuration.nas_secure_file_permissions == 'false':
            LOG.warning("The NAS file permissions mode will be 666 "
                        "(allowing other/world read & write access).")

    def _qemu_img_info(self, path, volume_name, force_share=True):
        return self._qemu_img_info_base(
            path, volume_name, self.configuration.quobyte_mount_point_base,
            force_share=True)

    @utils.synchronized('quobyte', external=False)
    def create_cloned_volume(self, volume, src_vref):
        """Creates a clone of the specified volume."""
        return self._create_cloned_volume(volume, src_vref,
                                          src_vref.obj_context)

    @coordination.synchronized(
        '{self.driver_prefix}-{snapshot.id}-{volume.id}')
    def _create_volume_from_snapshot(self, volume, snapshot):
        """Creates a volume from a snapshot.

        Snapshot must not be the active snapshot. (offline)
        """

        LOG.debug('Creating volume %(vol)s from snapshot %(snap)s',
                  {'vol': volume.id, 'snap': snapshot.id})

        if snapshot.status not in ['available', 'backing-up']:
            msg = _('Snapshot status must be "available" or "backing-up" to '
                    'clone. But is: %(status)s') % {'status': snapshot.status}

            raise exception.InvalidSnapshot(msg)

        self._ensure_shares_mounted()

        volume.provider_location = self._find_share(volume)

        self._copy_volume_from_snapshot(snapshot,
                                        volume,
                                        volume.size)

        return {'provider_location': volume.provider_location}

    @utils.synchronized('quobyte', external=False)
    def create_volume(self, volume):
        return super(QuobyteDriver, self).create_volume(volume)

    @utils.synchronized('quobyte', external=False)
    def create_volume_from_snapshot(self, volume, snapshot):
        return self._create_volume_from_snapshot(volume, snapshot)

    @coordination.synchronized('{self.driver_prefix}-{volume.id}')
    def _copy_volume_from_snapshot(self, snapshot, volume, volume_size):
        """Copy data from snapshot to destination volume.

        This is done with a qemu-img convert to raw/qcow2 from the snapshot
        qcow2. If the quobyte_volume_from_snapshot_cache is active the result
        is written into the cache and all volumes created from this
        snapshot id are created directly from the cache.
        """

        LOG.debug("snapshot: %(snap)s, volume: %(vol)s, ",
                  {'snap': snapshot.id,
                   'vol': volume.id,
                   'size': volume_size})
        info_path = self._local_path_volume_info(snapshot.volume)
        snap_info = self._read_info_file(info_path)
        vol_path = self._local_volume_dir(snapshot.volume)
        forward_file = snap_info[snapshot.id]
        forward_path = os.path.join(vol_path, forward_file)

        self._ensure_shares_mounted()
        # Find the file which backs this file, which represents the point
        # when this snapshot was created.
        img_info = self._qemu_img_info(forward_path, snapshot.volume.name)
        path_to_snap_img = os.path.join(vol_path, img_info.backing_file)

        path_to_new_vol = self._local_path_volume(volume)
        path_to_cached_vol = self._local_volume_from_snap_cache_path(snapshot)

        LOG.debug("will copy from snapshot at %s", path_to_snap_img)

        if self.configuration.quobyte_qcow2_volumes:
            out_format = 'qcow2'
        else:
            out_format = 'raw'

        if not self.configuration.quobyte_volume_from_snapshot_cache:
            LOG.debug("Creating direct copy from snapshot")
            image_utils.convert_image(path_to_snap_img,
                                      path_to_new_vol,
                                      out_format,
                                      run_as_root=self._execute_as_root)
        else:
            # create the volume via volume cache
            if not os.access(path_to_cached_vol, os.F_OK):
                LOG.debug("Caching volume %(volpath)s from snapshot.",
                          {'volpath': path_to_cached_vol})
                image_utils.convert_image(path_to_snap_img,
                                          path_to_cached_vol,
                                          out_format,
                                          run_as_root=self._execute_as_root)
                if self.configuration.quobyte_overlay_volumes:
                    # NOTE(kaisers): Create a parent symlink to track the
                    # existence of the parent
                    os.symlink(path_to_snap_img, path_to_cached_vol
                               + '.parent-' + snapshot.id)
            if self.configuration.quobyte_overlay_volumes:
                self._create_overlay_volume_from_snapshot(volume,
                                                          snapshot,
                                                          volume_size,
                                                          out_format)
            else:
                # Copy volume from cache
                LOG.debug("Copying volume %(volpath)s from cache",
                          {'volpath': path_to_new_vol})
                shutil.copyfile(path_to_cached_vol, path_to_new_vol)
                # Note(kaisers): As writes beyond EOF are sequentialized with
                # FUSE we call fallocate here to optimize performance:
                self._fallocate_file(path_to_new_vol, volume_size)
        self._set_rw_permissions(path_to_new_vol)

    def _create_overlay_volume_from_snapshot(self, volume, snapshot,
                                             volume_size, out_format):
        """Creates an overlay volume based on a parent in the cache

        Besides the overlay volume this also creates a softlink in the cache
        that  links to the child volume file of the cached volume. This can
        be used to track the cached volumes child volume and marks the fact
        that this child still exists. The softlink is deleted when
        the child is deleted.
        """
        rel_path = os.path.join(
            self.QUOBYTE_VOLUME_SNAP_CACHE_DIR_NAME, snapshot.id)
        command = ['qemu-img', 'create', '-f', 'qcow2', '-o',
                   'backing_file=%s,backing_fmt=%s' %
                   (rel_path, out_format), self._local_path_volume(volume),
                   "%dG" % volume_size]
        self._execute(*command, run_as_root=self._execute_as_root)
        os.symlink(self._local_path_volume(volume),
                   self._local_volume_from_snap_cache_path(snapshot)
                   + '.child-' + volume.id)

    @coordination.synchronized('{self.driver_prefix}-{volume.id}')
    def delete_volume(self, volume):
        """Deletes a logical volume."""

        if not volume.provider_location:
            LOG.warning('Volume %s does not have provider_location '
                        'specified, skipping', volume.name)
            return

        self._ensure_share_mounted(volume.provider_location)

        volume_dir = self._local_volume_dir(volume)
        active_image = self.get_active_image_from_info(volume)
        mounted_path = os.path.join(volume_dir, active_image)
        if os.access(self.local_path(volume), os.F_OK):
            img_info = self._qemu_img_info(self.local_path(volume),
                                           volume.name)
            if (img_info.backing_file and
                    (self.QUOBYTE_VOLUME_SNAP_CACHE_DIR_NAME in
                        img_info.backing_file)):
                # This is an overlay volume, call cache cleanup
                self._remove_from_vol_cache(img_info.backing_file,
                                            ".child-" + volume.id, volume)

        self._execute('rm', '-f', mounted_path,
                      run_as_root=self._execute_as_root)

        # If an exception (e.g. timeout) occurred during delete_snapshot, the
        # base volume may linger around, so just delete it if it exists
        base_volume_path = self._local_path_volume(volume)
        fileutils.delete_if_exists(base_volume_path)

        info_path = self._local_path_volume_info(volume)
        fileutils.delete_if_exists(info_path)

    @utils.synchronized('quobyte', external=False)
    def create_snapshot(self, snapshot):
        """Apply locking to the create snapshot operation."""

        return self._create_snapshot(snapshot)

    @utils.synchronized('quobyte', external=False)
    def initialize_connection(self, volume, connector):
        """Allow connection to connector and return connection info."""

        # Find active qcow2 file
        active_file = self.get_active_image_from_info(volume)
        path = '%s/%s/%s' % (self.configuration.quobyte_mount_point_base,
                             self._get_hash_str(volume.provider_location),
                             active_file)

        data = {'export': volume.provider_location,
                'name': active_file}
        if volume.provider_location in self.shares:
            data['options'] = self.shares[volume.provider_location]

        # Test file for raw vs. qcow2 format
        info = self._qemu_img_info(path, volume.name)
        data['format'] = info.file_format
        if data['format'] not in ['raw', 'qcow2']:
            msg = _('%s must be a valid raw or qcow2 image.') % path
            raise exception.InvalidVolume(msg)

        return {
            'driver_volume_type': 'quobyte',
            'data': data,
            'mount_point_base': self.configuration.quobyte_mount_point_base
        }

    @utils.synchronized('quobyte', external=False)
    def copy_volume_to_image(self, context, volume, image_service, image_meta):
        self._copy_volume_to_image(context, volume, image_service,
                                   image_meta)

    @utils.synchronized('quobyte', external=False)
    def extend_volume(self, volume, size_gb):
        if self._is_volume_attached(volume):
            # NOTE(kaisers): no attached extensions until #1870367 is fixed
            msg = (_("Cannot extend volume %s while it is attached.")
                   % volume['id'])
            raise exception.ExtendVolumeError(msg)

        volume_path = self.local_path(volume)

        info = self._qemu_img_info(volume_path, volume.name)
        backing_fmt = info.file_format

        if backing_fmt not in ['raw', 'qcow2']:
            msg = _('Unrecognized backing format: %s')
            raise exception.InvalidVolume(msg % backing_fmt)

        # qemu-img can resize both raw and qcow2 files
        active_path = os.path.join(
            self._get_mount_point_for_share(volume.provider_location),
            self.get_active_image_from_info(volume))
        image_utils.resize_image(active_path, size_gb)

    def _do_create_volume(self, volume):
        """Create a volume on given Quobyte volume.

        :param volume: volume reference
        """
        volume_path = self.local_path(volume)
        volume_size = volume.size

        if self.configuration.quobyte_qcow2_volumes:
            self._create_qcow2_file(volume_path, volume_size)
        else:
            if self.configuration.quobyte_sparsed_volumes:
                self._create_sparsed_file(volume_path, volume_size)
            else:
                self._create_regular_file(volume_path, volume_size)

        self._set_rw_permissions_for_all(volume_path)

    def _load_shares_config(self, share_file=None):
        """Put 'quobyte_volume_url' into the 'shares' list.

        :param share_file: string, Not used because the user has to specify
                                   the Quobyte volume directly.
        """
        self.shares = {}

        url = self._strip_qb_protocol(self.configuration.quobyte_volume_url)

        self.shares[url] = None  # None = No extra mount options.

        LOG.debug("Quobyte Volume URL set to: %s", self.shares)

    def _ensure_share_mounted(self, quobyte_volume):
        """Mount Quobyte volume.

        :param quobyte_volume: string
        """
        mount_path = self._get_mount_point_for_share(quobyte_volume)
        # NOTE(kaisers): Always use the currently configured volume url
        self._mount_quobyte(
            self._strip_qb_protocol(self.configuration.quobyte_volume_url),
            mount_path, ensure=True)

    @utils.synchronized('quobyte_ensure', external=False)
    def _ensure_shares_mounted(self):
        """Mount the Quobyte volume.

        Used for example by RemoteFsDriver._update_volume_stats
        """
        self._mounted_shares = []

        self._load_shares_config()

        for share in self.shares:
            try:
                self._ensure_share_mounted(share)
                self._mounted_shares.append(share)
            except Exception as exc:
                LOG.warning('Exception during mounting %s', exc)

        LOG.debug('Available shares %s', self._mounted_shares)

    def _find_share(self, volume):
        """Returns the mounted Quobyte volume.

        Multiple shares are not supported because the virtualization of
        multiple storage devices is taken care of at the level of Quobyte USP.

        For different types of volumes e.g., SSD vs. rotating disks, use
        multiple backends in Cinder.

        :param volume: the volume to be created.
        """

        if not self._mounted_shares:
            raise exception.NotFound()

        assert len(self._mounted_shares) == 1, 'There must be exactly' \
            ' one Quobyte volume.'
        target_volume = self._mounted_shares[0]

        LOG.debug('Selected %s as target Quobyte volume.', target_volume)

        return target_volume

    def _get_mount_point_for_share(self, quobyte_volume):
        """Return mount point for Quobyte volume.

        :param quobyte_volume: Example: storage-host/openstack-volumes
        """
        return os.path.join(self.configuration.quobyte_mount_point_base,
                            self._get_hash_str(quobyte_volume))

    # open() wrapper to mock reading from /proc/mount.
    @staticmethod
    def read_proc_mount():  # pragma: no cover
        return open('/proc/mounts')

    def _mount_quobyte(self, quobyte_volume, mount_path, ensure=False):
        """Mount Quobyte volume to mount path."""
        mounted = False
        with QuobyteDriver.read_proc_mount() as proc_mount:
            for line in proc_mount:
                if line.split()[1] == mount_path:
                    mounted = True
                    break

        if mounted:
            try:
                os.stat(mount_path)
            except OSError as exc:
                if exc.errno == errno.ENOTCONN:
                    mounted = False
                    try:
                        LOG.info('Fixing previous mount %s which was not'
                                 ' unmounted correctly.', mount_path)
                        self._execute('umount.quobyte', mount_path,
                                      run_as_root=self._execute_as_root)
                    except processutils.ProcessExecutionError as exc:
                        LOG.warning("Failed to unmount previous mount: "
                                    "%s", exc)
                else:
                    # TODO(quobyte): Extend exc analysis in here?
                    LOG.warning("Unknown error occurred while checking "
                                "mount point: %s Trying to continue.",
                                exc)

        if not mounted:
            if not os.path.isdir(mount_path):
                fileutils.ensure_tree(mount_path)

            command = ['mount.quobyte', '--disable-xattrs',
                       quobyte_volume, mount_path]
            if self.configuration.quobyte_client_cfg:
                command.extend(['-c', self.configuration.quobyte_client_cfg])

            try:
                LOG.info('Mounting volume: %s ...', quobyte_volume)
                self._execute(*command, run_as_root=self._execute_as_root)
                LOG.info('Mounting volume: %s succeeded', quobyte_volume)
                mounted = True
            except processutils.ProcessExecutionError as exc:
                if ensure and 'already mounted' in exc.stderr:
                    LOG.warning("%s is already mounted", quobyte_volume)
                    mounted = True
                else:
                    raise

        if mounted:
            self._validate_volume(mount_path)
            if self.configuration.quobyte_volume_from_snapshot_cache:
                self._ensure_volume_from_snap_cache(mount_path)

    def _validate_volume(self, mount_path):
        """Runs a number of tests on the expect Quobyte mount"""
        partitions = psutil.disk_partitions(all=True)
        for p in partitions:
            if mount_path == p.mountpoint:
                if (p.device.startswith("quobyte@") or
                        (p.fstype == "fuse.quobyte")):
                    try:
                        statresult = os.stat(mount_path)
                        if statresult.st_size == 0:
                            # client looks healthy
                            if not os.access(mount_path,
                                             os.W_OK | os.X_OK):
                                LOG.warning("Volume is not writable. "
                                            "Please broaden the file"
                                            " permissions."
                                            " Mount: %s",
                                            mount_path)
                            return  # we're happy here
                        else:
                            msg = (_("The mount %(mount_path)s is not a "
                                     "valid Quobyte volume. Stale mount?")
                                   % {'mount_path': mount_path})
                        raise exception.VolumeDriverException(msg)
                    except Exception as exc:
                        msg = (_("The mount %(mount_path)s is not a valid"
                                 " Quobyte volume. Error: %(exc)s . "
                                 " Possibly a Quobyte client crash?")
                               % {'mount_path': mount_path, 'exc': exc})
                        raise exception.VolumeDriverException(msg)
                else:
                    msg = (_("The mount %(mount_path)s is not a valid"
                             " Quobyte volume according to partition list.")
                           % {'mount_path': mount_path})
                    raise exception.VolumeDriverException(msg)
        msg = (_("No matching Quobyte mount entry for %(mount_path)s"
                 " could be found for validation in partition list.")
               % {'mount_path': mount_path})
        raise exception.VolumeDriverException(msg)



# Copyright 2019 Nexenta Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from cinder.volume import configuration as conf

DEFAULT_ISCSI_PORT = 3260
DEFAULT_HOST_GROUP = 'all'
DEFAULT_TARGET_GROUP = 'all'

NEXENTA_EDGE_OPTS = [
    cfg.StrOpt('nexenta_nbd_symlinks_dir',
               default='/dev/disk/by-path',
               help='NexentaEdge logical path of directory to store symbolic '
                    'links to NBDs'),
    cfg.StrOpt('nexenta_rest_user',
               default='admin',
               help='User name to connect to NexentaEdge.'),
    cfg.StrOpt('nexenta_rest_password',
               default='nexenta',
               help='Password to connect to NexentaEdge.',
               secret=True),
    cfg.StrOpt('nexenta_lun_container',
               default='',
               help='NexentaEdge logical path of bucket for LUNs'),
    cfg.StrOpt('nexenta_iscsi_service',
               default='',
               help='NexentaEdge iSCSI service name'),
    cfg.StrOpt('nexenta_client_address',
               deprecated_for_removal=True,
               deprecated_reason='iSCSI target address should now be set using'
                                 ' the common param target_ip_address.',
               default='',
               help='NexentaEdge iSCSI Gateway client '
               'address for non-VIP service'),
    cfg.IntOpt('nexenta_iops_limit',
               default=0,
               help='NexentaEdge iSCSI LUN object IOPS limit'),
    cfg.IntOpt('nexenta_chunksize',
               default=32768,
               help='NexentaEdge iSCSI LUN object chunk size'),
    cfg.IntOpt('nexenta_replication_count',
               default=3,
               help='NexentaEdge iSCSI LUN object replication count.'),
    cfg.BoolOpt('nexenta_encryption',
                default=False,
                help='Defines whether NexentaEdge iSCSI LUN object '
                     'has encryption enabled.')
]

NEXENTA_CONNECTION_OPTS = [
    cfg.StrOpt('nexenta_host',
               default='',
               help='IP address of NexentaStor Appliance'),
    cfg.StrOpt('nexenta_rest_address',
               deprecated_for_removal=True,
               deprecated_reason='Rest address should now be set using '
                                 'the common param depending on driver type, '
                                 'san_ip or nas_host',
               default='',
               help='IP address of NexentaStor management REST API endpoint'),
    cfg.IntOpt('nexenta_rest_port',
               deprecated_for_removal=True,
               deprecated_reason='Rest address should now be set using '
                                 'the common param san_api_port.',
               default=0,
               help='HTTP(S) port to connect to NexentaStor management '
                    'REST API server. If it is equal zero, 8443 for '
                    'HTTPS and 8080 for HTTP is used'),
    cfg.StrOpt('nexenta_rest_protocol',
               default='auto',
               choices=['http', 'https', 'auto'],
               help='Use http or https for NexentaStor management '
                    'REST API connection (default auto)'),
    cfg.FloatOpt('nexenta_rest_connect_timeout',
                 default=30,
                 help='Specifies the time limit (in seconds), within '
                      'which the connection to NexentaStor management '
                      'REST API server must be established'),
    cfg.FloatOpt('nexenta_rest_read_timeout',
                 default=300,
                 help='Specifies the time limit (in seconds), '
                      'within which NexentaStor management '
                      'REST API server must send a response'),
    cfg.FloatOpt('nexenta_rest_backoff_factor',
                 default=0.5,
                 help='Specifies the backoff factor to apply '
                      'between connection attempts to NexentaStor '
                      'management REST API server'),
    cfg.IntOpt('nexenta_rest_retry_count',
               default=3,
               help='Specifies the number of times to repeat NexentaStor '
                    'management REST API call in case of connection errors '
                    'and NexentaStor appliance EBUSY or ENOENT errors'),
    cfg.BoolOpt('nexenta_use_https',
                default=True,
                help='Use HTTP secure protocol for NexentaStor '
                     'management REST API connections'),
    cfg.BoolOpt('nexenta_lu_writebackcache_disabled',
                default=False,
                help='Postponed write to backing store or not'),
    cfg.StrOpt('nexenta_user',
               deprecated_for_removal=True,
               deprecated_reason='Common user parameters should be used '
                                 'depending on the driver type: '
                                 'san_login or nas_login',
               default='admin',
               help='User name to connect to NexentaStor '
                    'management REST API server'),
    cfg.StrOpt('nexenta_password',
               deprecated_for_removal=True,
               deprecated_reason='Common password parameters should be used '
                                 'depending on the driver type: '
                                 'san_password or nas_password',
               default='nexenta',
               help='Password to connect to NexentaStor '
                    'management REST API server',
               secret=True)
]

NEXENTA_ISCSI_OPTS = [
    cfg.StrOpt('nexenta_iscsi_target_portal_groups',
               default='',
               help='NexentaStor target portal groups'),
    cfg.StrOpt('nexenta_iscsi_target_portals',
               default='',
               help='Comma separated list of portals for NexentaStor5, in '
                    'format of IP1:port1,IP2:port2. Port is optional, '
                    'default=3260. Example: 10.10.10.1:3267,10.10.1.2'),
    cfg.StrOpt('nexenta_iscsi_target_host_group',
               default='all',
               help='Group of hosts which are allowed to access volumes'),
    cfg.IntOpt('nexenta_iscsi_target_portal_port',
               default=3260,
               help='Nexenta appliance iSCSI target portal port'),
    cfg.IntOpt('nexenta_luns_per_target',
               default=100,
               help='Amount of LUNs per iSCSI target'),
    cfg.StrOpt('nexenta_volume',
               default='cinder',
               help='NexentaStor pool name that holds all volumes'),
    cfg.StrOpt('nexenta_target_prefix',
               default='iqn.1986-03.com.sun:02:cinder',
               help='iqn prefix for NexentaStor iSCSI targets'),
    cfg.StrOpt('nexenta_target_group_prefix',
               default='cinder',
               help='Prefix for iSCSI target groups on NexentaStor'),
    cfg.StrOpt('nexenta_host_group_prefix',
               default='cinder',
               help='Prefix for iSCSI host groups on NexentaStor'),
    cfg.StrOpt('nexenta_volume_group',
               default='iscsi',
               help='Volume group for NexentaStor5 iSCSI'),
]

NEXENTA_NFS_OPTS = [
    cfg.StrOpt('nexenta_shares_config',
               default='/etc/cinder/nfs_shares',
               help='File with the list of available nfs shares'),
    cfg.StrOpt('nexenta_mount_point_base',
               default='$state_path/mnt',
               help='Base directory that contains NFS share mount points'),
    cfg.BoolOpt('nexenta_sparsed_volumes',
                default=True,
                help='Enables or disables the creation of volumes as '
                     'sparsed files that take no space. If disabled '
                     '(False), volume is created as a regular file, '
                     'which takes a long time.'),
    cfg.BoolOpt('nexenta_qcow2_volumes',
                default=False,
                help='Create volumes as QCOW2 files rather than raw files'),
    cfg.BoolOpt('nexenta_nms_cache_volroot',
                default=True,
                help=('If set True cache NexentaStor appliance volroot option '
                      'value.'))
]

NEXENTA_DATASET_OPTS = [
    cfg.StrOpt('nexenta_dataset_compression',
               default='on',
               choices=['on', 'off', 'gzip', 'gzip-1', 'gzip-2', 'gzip-3',
                        'gzip-4', 'gzip-5', 'gzip-6', 'gzip-7', 'gzip-8',
                        'gzip-9', 'lzjb', 'zle', 'lz4'],
               help='Compression value for new ZFS folders.'),
    cfg.StrOpt('nexenta_dataset_dedup',
               default='off',
               choices=['on', 'off', 'sha256', 'verify', 'sha256, verify'],
               help='Deduplication value for new ZFS folders.'),
    cfg.StrOpt('nexenta_folder',
               default='',
               help='A folder where cinder created datasets will reside.'),
    cfg.StrOpt('nexenta_dataset_description',
               default='',
               help='Human-readable description for the folder.'),
    cfg.IntOpt('nexenta_blocksize',
               default=4096,
               help='Block size for datasets'),
    cfg.IntOpt('nexenta_ns5_blocksize',
               default=32,
               help='Block size for datasets'),
    cfg.BoolOpt('nexenta_sparse',
                default=False,
                help='Enables or disables the creation of sparse datasets'),
    cfg.StrOpt('nexenta_origin_snapshot_template',
               default='origin-snapshot-%s',
               help='Template string to generate origin name of clone'),
    cfg.StrOpt('nexenta_group_snapshot_template',
               default='group-snapshot-%s',
               help='Template string to generate group snapshot name')
]

NEXENTA_RRMGR_OPTS = [
    cfg.IntOpt('nexenta_rrmgr_compression',
               default=0,
               help=('Enable stream compression, level 1..9. 1 - gives best '
                     'speed; 9 - gives best compression.')),
    cfg.IntOpt('nexenta_rrmgr_tcp_buf_size',
               default=4096,
               help='TCP Buffer size in KiloBytes.'),
    cfg.IntOpt('nexenta_rrmgr_connections',
               default=2,
               help='Number of TCP connections.'),
]

CONF = cfg.CONF
CONF.register_opts(NEXENTA_CONNECTION_OPTS, group=conf.SHARED_CONF_GROUP)
CONF.register_opts(NEXENTA_ISCSI_OPTS, group=conf.SHARED_CONF_GROUP)
CONF.register_opts(NEXENTA_DATASET_OPTS, group=conf.SHARED_CONF_GROUP)
CONF.register_opts(NEXENTA_NFS_OPTS, group=conf.SHARED_CONF_GROUP)
CONF.register_opts(NEXENTA_RRMGR_OPTS, group=conf.SHARED_CONF_GROUP)
CONF.register_opts(NEXENTA_EDGE_OPTS, group=conf.SHARED_CONF_GROUP)



#!/usr/bin/env python

# Copyright 2012 - 2013 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import os
import sys

import rtslib_fb

from cinder import i18n
from cinder.i18n import _

i18n.enable_lazy()


class RtstoolError(Exception):
    pass


class RtstoolImportError(RtstoolError):
    pass


def create(backing_device, name, userid, password, iser_enabled,
           initiator_iqns=None, portals_ips=None, portals_port=3260):
    # List of IPS that will not raise an error when they fail binding.
    # Originally we will fail on all binding errors.
    ips_allow_fail = ()

    try:
        rtsroot = rtslib_fb.root.RTSRoot()
    except rtslib_fb.utils.RTSLibError:
        print(_('Ensure that configfs is mounted at /sys/kernel/config.'))
        raise

    # Look to see if BlockStorageObject already exists
    for x in rtsroot.storage_objects:
        if x.name == name:
            # Already exists, use this one
            return

    so_new = rtslib_fb.BlockStorageObject(name=name,
                                          dev=backing_device)

    target_new = rtslib_fb.Target(rtslib_fb.FabricModule('iscsi'), name,
                                  'create')

    tpg_new = rtslib_fb.TPG(target_new, mode='create')
    tpg_new.set_attribute('authentication', '1')

    lun_new = rtslib_fb.LUN(tpg_new, storage_object=so_new)

    if initiator_iqns:
        initiator_iqns = initiator_iqns.strip(' ')
        for i in initiator_iqns.split(','):
            acl_new = rtslib_fb.NodeACL(tpg_new, i, mode='create')
            acl_new.chap_userid = userid
            acl_new.chap_password = password

            rtslib_fb.MappedLUN(acl_new, lun_new.lun, lun_new.lun)

    tpg_new.enable = 1

    # If no ips are given we'll bind to all IPv4 and v6
    if not portals_ips:
        portals_ips = ('0.0.0.0', '[::0]')
        # TODO(emh): Binding to IPv6 fails sometimes -- let pass for now.
        ips_allow_fail = ('[::0]',)

    for ip in portals_ips:
        try:
            # rtslib expects IPv6 addresses to be surrounded by brackets
            portal = rtslib_fb.NetworkPortal(tpg_new, _canonicalize_ip(ip),
                                             portals_port, mode='any')
        except rtslib_fb.utils.RTSLibError:
            raise_exc = ip not in ips_allow_fail
            msg_type = 'Error' if raise_exc else 'Warning'
            print(_('%(msg_type)s: creating NetworkPortal: ensure port '
                  '%(port)d on ip %(ip)s is not in use by another service.')
                  % {'msg_type': msg_type, 'port': portals_port, 'ip': ip})
            if raise_exc:
                raise
        else:
            try:
                if iser_enabled == 'True':
                    portal.iser = True
            except rtslib_fb.utils.RTSLibError:
                print(_('Error enabling iSER for NetworkPortal: please ensure '
                        'that RDMA is supported on your iSCSI port %(port)d '
                        'on ip %(ip)s.') % {'port': portals_port, 'ip': ip})
                raise


def _lookup_target(target_iqn, initiator_iqn):
    try:
        rtsroot = rtslib_fb.root.RTSRoot()
    except rtslib_fb.utils.RTSLibError:
        print(_('Ensure that configfs is mounted at /sys/kernel/config.'))
        raise

    # Look for the target
    for t in rtsroot.targets:
        if t.wwn == target_iqn:
            return t
    raise RtstoolError(_('Could not find target %s') % target_iqn)


def add_initiator(target_iqn, initiator_iqn, userid, password):
    target = _lookup_target(target_iqn, initiator_iqn)
    tpg = next(target.tpgs)  # get the first one
    for acl in tpg.node_acls:
        # See if this ACL configuration already exists
        if acl.node_wwn.lower() == initiator_iqn.lower():
            # No further action required
            return

    acl_new = rtslib_fb.NodeACL(tpg, initiator_iqn, mode='create')
    acl_new.chap_userid = userid
    acl_new.chap_password = password

    rtslib_fb.MappedLUN(acl_new, 0, tpg_lun=0)


def delete_initiator(target_iqn, initiator_iqn):
    target = _lookup_target(target_iqn, initiator_iqn)
    tpg = next(target.tpgs)  # get the first one
    for acl in tpg.node_acls:
        if acl.node_wwn.lower() == initiator_iqn.lower():
            acl.delete()
            return

    print(_('delete_initiator: %s ACL not found. Continuing.') % initiator_iqn)
    # Return successfully.


def get_targets():
    rtsroot = rtslib_fb.root.RTSRoot()
    for x in rtsroot.targets:
        print(x.wwn)


def delete(iqn):
    rtsroot = rtslib_fb.root.RTSRoot()
    for x in rtsroot.targets:
        if x.wwn == iqn:
            x.delete()
            break

    for x in rtsroot.storage_objects:
        if x.name == iqn:
            x.delete()
            break


def verify_rtslib():
    for member in ['BlockStorageObject', 'FabricModule', 'LUN',
                   'MappedLUN', 'NetworkPortal', 'NodeACL', 'root',
                   'Target', 'TPG']:
        if not hasattr(rtslib_fb, member):
            raise RtstoolImportError(_("rtslib_fb is missing member %s: You "
                                       "may need a newer python-rtslib-fb.") %
                                     member)


def usage():
    print("Usage:")
    print(sys.argv[0] +
          " create [device] [name] [userid] [password] [iser_enabled]" +
          " <initiator_iqn,iqn2,iqn3,...> [-a<IP1,IP2,...>] [-pPORT]")
    print(sys.argv[0] +
          " add-initiator [target_iqn] [userid] [password] [initiator_iqn]")
    print(sys.argv[0] +
          " delete-initiator [target_iqn] [initiator_iqn]")
    print(sys.argv[0] + " get-targets")
    print(sys.argv[0] + " delete [iqn]")
    print(sys.argv[0] + " verify")
    print(sys.argv[0] + " save [path_to_file]")
    sys.exit(1)


def save_to_file(destination_file):
    rtsroot = rtslib_fb.root.RTSRoot()
    try:
        # If default destination use rtslib default save file
        if not destination_file:
            destination_file = rtslib_fb.root.default_save_file
            path_to_file = os.path.dirname(destination_file)

            # NOTE(geguileo): With default file we ensure path exists and
            # create it if doesn't.
            # Cinder's LIO target helper runs this as root, so it will have no
            # problem creating directory /etc/target.
            # If run manually from the command line without being root you will
            # get an error, same as when creating and removing targets.
            if not os.path.exists(path_to_file):
                os.makedirs(path_to_file, 0o755)

    except OSError as exc:
        raise RtstoolError(_('targetcli not installed and could not create '
                             'default directory (%(default_path)s): %(exc)s') %
                           {'default_path': path_to_file, 'exc': exc})
    try:
        rtsroot.save_to_file(destination_file)
    except (OSError, IOError) as exc:
        raise RtstoolError(_('Could not save configuration to %(file_path)s: '
                             '%(exc)s') %
                           {'file_path': destination_file, 'exc': exc})


def restore_from_file(configuration_file):
    rtsroot = rtslib_fb.root.RTSRoot()
    # If configuration file is None, use rtslib default save file.
    if not configuration_file:
        configuration_file = rtslib_fb.root.default_save_file

    try:
        rtsroot.restore_from_file(configuration_file)
    except (OSError, IOError) as exc:
        raise RtstoolError(_('Could not restore configuration file '
                             '%(file_path)s: %(exc)s'),
                           {'file_path': configuration_file, 'exc': exc})


def parse_optional_create(argv):
    optional_args = {}

    for arg in argv:
        if arg.startswith('-a'):
            ips = [ip for ip in arg[2:].split(',') if ip]
            if not ips:
                usage()
            optional_args['portals_ips'] = ips
        elif arg.startswith('-p'):
            try:
                optional_args['portals_port'] = int(arg[2:])
            except ValueError:
                usage()
        else:
            optional_args['initiator_iqns'] = arg
    return optional_args


def _canonicalize_ip(ip):
    if ip.startswith('[') or "." in ip:
        return ip
    return "[" + ip + "]"


def main(argv=None):
    if argv is None:
        argv = sys.argv

    if len(argv) < 2:
        usage()

    if argv[1] == 'create':
        if len(argv) < 7:
            usage()

        if len(argv) > 10:
            usage()

        backing_device = argv[2]
        name = argv[3]
        userid = argv[4]
        password = argv[5]
        iser_enabled = argv[6]

        if len(argv) > 7:
            optional_args = parse_optional_create(argv[7:])
        else:
            optional_args = {}

        create(backing_device, name, userid, password, iser_enabled,
               **optional_args)

    elif argv[1] == 'add-initiator':
        if len(argv) < 6:
            usage()

        target_iqn = argv[2]
        userid = argv[3]
        password = argv[4]
        initiator_iqn = argv[5]

        add_initiator(target_iqn, initiator_iqn, userid, password)

    elif argv[1] == 'delete-initiator':
        if len(argv) < 4:
            usage()

        target_iqn = argv[2]
        initiator_iqn = argv[3]

        delete_initiator(target_iqn, initiator_iqn)

    elif argv[1] == 'get-targets':
        get_targets()

    elif argv[1] == 'delete':
        if len(argv) < 3:
            usage()

        iqn = argv[2]
        delete(iqn)

    elif argv[1] == 'verify':
        # This is used to verify that this script can be called by cinder,
        # and that rtslib_fb is new enough to work.
        verify_rtslib()
        return 0

    elif argv[1] == 'save':
        if len(argv) > 3:
            usage()

        destination_file = argv[2] if len(argv) > 2 else None
        save_to_file(destination_file)
        return 0

    elif argv[1] == 'restore':
        if len(argv) > 3:
            usage()

        configuration_file = argv[2] if len(argv) > 2 else None
        restore_from_file(configuration_file)
        return 0

    else:
        usage()

    return 0



# Copyright (c) 2011 X.commerce, a business unit of eBay Inc.
# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Defines interface for DB access.

Functions in this module are imported into the cinder.db namespace. Call these
functions from cinder.db namespace, not the cinder.db.api namespace.

All functions in this module return objects that implement a dictionary-like
interface. Currently, many of these objects are sqlalchemy objects that
implement a dictionary interface. However, a future goal is to have all of
these objects be simple dictionaries.


**Related Flags**

:connection:  string specifying the sqlalchemy connection to use, like:
              `sqlite:///var/lib/cinder/cinder.sqlite`.

:enable_new_services:  when adding a new service to the database, is it in the
                       pool of available hardware (Default: True)

"""

from oslo_config import cfg
from oslo_db import api as oslo_db_api
from oslo_db import options as db_options

from cinder.api import common
from cinder.common import constants
from cinder.i18n import _

db_opts = [
    cfg.BoolOpt('enable_new_services',
                default=True,
                help='Services to be added to the available pool on create'),
    cfg.StrOpt('volume_name_template',
               default='volume-%s',
               help='Template string to be used to generate volume names'),
    cfg.StrOpt('snapshot_name_template',
               default='snapshot-%s',
               help='Template string to be used to generate snapshot names'),
]

backup_opts = [
    cfg.StrOpt('backup_name_template',
               default='backup-%s',
               help='Template string to be used to generate backup names'),
]

CONF = cfg.CONF
CONF.register_opts(db_opts)
CONF.register_opts(backup_opts)
db_options.set_defaults(CONF)

_BACKEND_MAPPING = {'sqlalchemy': 'cinder.db.sqlalchemy.api'}


IMPL = oslo_db_api.DBAPI.from_config(conf=CONF,
                                     backend_mapping=_BACKEND_MAPPING,
                                     lazy=True)

# The maximum value a signed INT type may have
MAX_INT = constants.DB_MAX_INT


###################

def dispose_engine():
    """Force the engine to establish new connections."""

    # FIXME(jdg): When using sqlite if we do the dispose
    # we seem to lose our DB here.  Adding this check
    # means we don't do the dispose, but we keep our sqlite DB
    # This likely isn't the best way to handle this

    if 'sqlite' not in IMPL.get_engine().name:
        return IMPL.dispose_engine()
    else:
        return


###################


def service_destroy(context, service_id):
    """Destroy the service or raise if it does not exist."""
    return IMPL.service_destroy(context, service_id)


def service_get(context, service_id=None, backend_match_level=None, **filters):
    """Get a service that matches the criteria.

    A possible filter is is_up=True and it will filter nodes that are down.

    :param service_id: Id of the service.
    :param filters: Filters for the query in the form of key/value.
    :param backend_match_level: 'pool', 'backend', or 'host' for host and
                                cluster filters (as defined in _filter_host
                                method)
    :raise ServiceNotFound: If service doesn't exist.
    """
    return IMPL.service_get(context, service_id, backend_match_level,
                            **filters)


def service_get_all(context, backend_match_level=None, **filters):
    """Get all services that match the criteria.

    A possible filter is is_up=True and it will filter nodes that are down,
    as well as host_or_cluster, that lets you look for services using both
    of these properties.

    :param filters: Filters for the query in the form of key/value arguments.
    :param backend_match_level: 'pool', 'backend', or 'host' for host and
                                cluster filters (as defined in _filter_host
                                method)
    """
    return IMPL.service_get_all(context, backend_match_level, **filters)


def service_create(context, values):
    """Create a service from the values dictionary."""
    return IMPL.service_create(context, values)


def service_update(context, service_id, values):
    """Set the given properties on an service and update it.

    Raises NotFound if service does not exist.
    """
    return IMPL.service_update(context, service_id, values)


def service_get_by_uuid(context, service_uuid):
    """Get a service by it's uuid.

    Return Service ref or raise if it does not exist.
    """
    return IMPL.service_get_by_uuid(context, service_uuid)


###############


def is_backend_frozen(context, host, cluster_name):
    """Check if a storage backend is frozen based on host and cluster_name."""
    return IMPL.is_backend_frozen(context, host, cluster_name)


###############


def cluster_get(context, id=None, is_up=None, get_services=False,
                services_summary=False, read_deleted='no',
                name_match_level=None, **filters):
    """Get a cluster that matches the criteria.

    :param id: Id of the cluster.
    :param is_up: Boolean value to filter based on the cluster's up status.
    :param get_services: If we want to load all services from this cluster.
    :param services_summary: If we want to load num_hosts and
                             num_down_hosts fields.
    :param read_deleted: Filtering based on delete status. Default value is
                         "no".
    :param name_match_level: 'pool', 'backend', or 'host' for name filter (as
                             defined in _filter_host method)
    :param filters: Field based filters in the form of key/value.
    :raise ClusterNotFound: If cluster doesn't exist.
    """
    return IMPL.cluster_get(context, id, is_up, get_services, services_summary,
                            read_deleted, name_match_level, **filters)


def cluster_get_all(context, is_up=None, get_services=False,
                    services_summary=False, read_deleted='no',
                    name_match_level=None, **filters):
    """Get all clusters that match the criteria.

    :param is_up: Boolean value to filter based on the cluster's up status.
    :param get_services: If we want to load all services from this cluster.
    :param services_summary: If we want to load num_hosts and
                             num_down_hosts fields.
    :param read_deleted: Filtering based on delete status. Default value is
                         "no".
    :param name_match_level: 'pool', 'backend', or 'host' for name filter (as
                             defined in _filter_host method)
    :param filters: Field based filters in the form of key/value.
    """
    return IMPL.cluster_get_all(context, is_up, get_services, services_summary,
                                read_deleted, name_match_level, **filters)


def cluster_create(context, values):
    """Create a cluster from the values dictionary."""
    return IMPL.cluster_create(context, values)


def cluster_update(context, id, values):
    """Set the given properties on an cluster and update it.

    Raises ClusterNotFound if cluster does not exist.
    """
    return IMPL.cluster_update(context, id, values)


def cluster_destroy(context, id):
    """Destroy the cluster or raise if it does not exist or has hosts.

    :raise ClusterNotFound: If cluster doesn't exist.
    """
    return IMPL.cluster_destroy(context, id)


###############


def volume_attach(context, values):
    """Attach a volume."""
    return IMPL.volume_attach(context, values)


def volume_attached(context, volume_id, instance_id, host_name, mountpoint,
                    attach_mode='rw', mark_attached=True):
    """Ensure that a volume is set as attached."""
    return IMPL.volume_attached(context, volume_id, instance_id, host_name,
                                mountpoint, attach_mode, mark_attached)


def volume_create(context, values):
    """Create a volume from the values dictionary."""
    return IMPL.volume_create(context, values)


def volume_data_get_for_host(context, host, count_only=False):
    """Get (volume_count, gigabytes) for project."""
    return IMPL.volume_data_get_for_host(context,
                                         host,
                                         count_only)


def volume_data_get_for_project(context, project_id, host=None):
    """Get (volume_count, gigabytes) for project."""
    return IMPL.volume_data_get_for_project(context, project_id, host=host)


def volume_destroy(context, volume_id):
    """Destroy the volume or raise if it does not exist."""
    return IMPL.volume_destroy(context, volume_id)


def volume_detached(context, volume_id, attachment_id):
    """Ensure that a volume is set as detached."""
    return IMPL.volume_detached(context, volume_id, attachment_id)


def volume_get(context, volume_id):
    """Get a volume or raise if it does not exist."""
    return IMPL.volume_get(context, volume_id)


def volume_get_all(context, marker=None, limit=None, sort_keys=None,
                   sort_dirs=None, filters=None, offset=None):
    """Get all volumes."""
    return IMPL.volume_get_all(context, marker, limit, sort_keys=sort_keys,
                               sort_dirs=sort_dirs, filters=filters,
                               offset=offset)


def calculate_resource_count(context, resource_type, filters):
    return IMPL.calculate_resource_count(context, resource_type, filters)


def volume_get_all_by_host(context, host, filters=None):
    """Get all volumes belonging to a host."""
    return IMPL.volume_get_all_by_host(context, host, filters=filters)


def volume_get_all_by_group(context, group_id, filters=None):
    """Get all volumes belonging to a consistency group."""
    return IMPL.volume_get_all_by_group(context, group_id, filters=filters)


def volume_get_all_by_generic_group(context, group_id, filters=None):
    """Get all volumes belonging to a generic volume group."""
    return IMPL.volume_get_all_by_generic_group(context, group_id,
                                                filters=filters)


def volume_get_all_by_project(context, project_id, marker, limit,
                              sort_keys=None, sort_dirs=None, filters=None,
                              offset=None):
    """Get all volumes belonging to a project."""
    return IMPL.volume_get_all_by_project(context, project_id, marker, limit,
                                          sort_keys=sort_keys,
                                          sort_dirs=sort_dirs,
                                          filters=filters,
                                          offset=offset)


def get_volume_summary(context, project_only, filters=None):
    """Get volume summary."""
    return IMPL.get_volume_summary(context, project_only, filters)


def volume_update(context, volume_id, values):
    """Set the given properties on a volume and update it.

    Raises NotFound if volume does not exist.

    """
    return IMPL.volume_update(context, volume_id, values)


def volumes_update(context, values_list):
    """Set the given properties on a list of volumes and update them.

    Raises NotFound if a volume does not exist.
    """
    return IMPL.volumes_update(context, values_list)


def volume_include_in_cluster(context, cluster, partial_rename=True,
                              **filters):
    """Include all volumes matching the filters into a cluster.

    When partial_rename is set we will not set the cluster_name with cluster
    parameter value directly, we'll replace provided cluster_name or host
    filter value with cluster instead.

    This is useful when we want to replace just the cluster name but leave
    the backend and pool information as it is.  If we are using cluster_name
    to filter, we'll use that same DB field to replace the cluster value and
    leave the rest as it is.  Likewise if we use the host to filter.

    Returns the number of volumes that have been changed.
    """
    return IMPL.volume_include_in_cluster(context, cluster, partial_rename,
                                          **filters)


def volume_attachment_update(context, attachment_id, values):
    return IMPL.volume_attachment_update(context, attachment_id, values)


def volume_attachment_get(context, attachment_id):
    return IMPL.volume_attachment_get(context, attachment_id)


def volume_attachment_get_all_by_volume_id(context, volume_id,
                                           session=None):
    return IMPL.volume_attachment_get_all_by_volume_id(context,
                                                       volume_id,
                                                       session)


def volume_attachment_get_all_by_host(context, host, filters=None):
    # FIXME(jdg): Not using filters
    return IMPL.volume_attachment_get_all_by_host(context, host)


def volume_attachment_get_all_by_instance_uuid(context,
                                               instance_uuid, filters=None):
    # FIXME(jdg): Not using filters
    return IMPL.volume_attachment_get_all_by_instance_uuid(context,
                                                           instance_uuid)


def volume_attachment_get_all(context, filters=None, marker=None, limit=None,
                              offset=None, sort_keys=None, sort_dirs=None):
    return IMPL.volume_attachment_get_all(context, filters, marker, limit,
                                          offset, sort_keys, sort_dirs)


def volume_attachment_get_all_by_project(context, project_id, filters=None,
                                         marker=None, limit=None, offset=None,
                                         sort_keys=None, sort_dirs=None):
    return IMPL.volume_attachment_get_all_by_project(context, project_id,
                                                     filters, marker, limit,
                                                     offset, sort_keys,
                                                     sort_dirs)


def attachment_destroy(context, attachment_id):
    """Destroy the attachment or raise if it does not exist."""
    return IMPL.attachment_destroy(context, attachment_id)


def volume_update_status_based_on_attachment(context, volume_id):
    """Update volume status according to attached instance id"""
    return IMPL.volume_update_status_based_on_attachment(context, volume_id)


def volume_has_snapshots_filter():
    return IMPL.volume_has_snapshots_filter()


def volume_has_undeletable_snapshots_filter():
    return IMPL.volume_has_undeletable_snapshots_filter()


def volume_has_snapshots_in_a_cgsnapshot_filter():
    return IMPL.volume_has_snapshots_in_a_cgsnapshot_filter()


def volume_has_attachments_filter():
    return IMPL.volume_has_attachments_filter()


def volume_qos_allows_retype(new_vol_type):
    return IMPL.volume_qos_allows_retype(new_vol_type)


def volume_has_other_project_snp_filter():
    return IMPL.volume_has_other_project_snp_filter()


####################


def snapshot_create(context, values):
    """Create a snapshot from the values dictionary."""
    return IMPL.snapshot_create(context, values)


def snapshot_destroy(context, snapshot_id):
    """Destroy the snapshot or raise if it does not exist."""
    return IMPL.snapshot_destroy(context, snapshot_id)


def snapshot_get(context, snapshot_id):
    """Get a snapshot or raise if it does not exist."""
    return IMPL.snapshot_get(context, snapshot_id)


def snapshot_get_all(context, filters=None, marker=None, limit=None,
                     sort_keys=None, sort_dirs=None, offset=None):
    """Get all snapshots."""
    return IMPL.snapshot_get_all(context, filters, marker, limit, sort_keys,
                                 sort_dirs, offset)


def snapshot_get_all_by_project(context, project_id, filters=None, marker=None,
                                limit=None, sort_keys=None, sort_dirs=None,
                                offset=None):
    """Get all snapshots belonging to a project."""
    return IMPL.snapshot_get_all_by_project(context, project_id, filters,
                                            marker, limit, sort_keys,
                                            sort_dirs, offset)


def snapshot_get_all_by_host(context, host, filters=None):
    """Get all snapshots belonging to a host.

    :param host: Include include snapshots only for specified host.
    :param filters: Filters for the query in the form of key/value.
    """
    return IMPL.snapshot_get_all_by_host(context, host, filters)


def snapshot_get_all_for_cgsnapshot(context, project_id):
    """Get all snapshots belonging to a cgsnapshot."""
    return IMPL.snapshot_get_all_for_cgsnapshot(context, project_id)


def snapshot_get_all_for_group_snapshot(context, group_snapshot_id):
    """Get all snapshots belonging to a group snapshot."""
    return IMPL.snapshot_get_all_for_group_snapshot(context, group_snapshot_id)


def snapshot_get_all_for_volume(context, volume_id):
    """Get all snapshots for a volume."""
    return IMPL.snapshot_get_all_for_volume(context, volume_id)


def snapshot_get_latest_for_volume(context, volume_id):
    """Get latest snapshot for a volume"""
    return IMPL.snapshot_get_latest_for_volume(context, volume_id)


def snapshot_update(context, snapshot_id, values):
    """Set the given properties on an snapshot and update it.

    Raises NotFound if snapshot does not exist.

    """
    return IMPL.snapshot_update(context, snapshot_id, values)


def snapshot_data_get_for_project(context, project_id, volume_type_id=None,
                                  host=None):
    """Get count and gigabytes used for snapshots for specified project."""
    return IMPL.snapshot_data_get_for_project(context,
                                              project_id,
                                              volume_type_id,
                                              host=host)


def snapshot_get_all_active_by_window(context, begin, end=None,
                                      project_id=None):
    """Get all the snapshots inside the window.

    Specifying a project_id will filter for a certain project.
    """
    return IMPL.snapshot_get_all_active_by_window(context, begin, end,
                                                  project_id)


def get_snapshot_summary(context, project_only, filters=None):
    """Get snapshot summary."""
    return IMPL.get_snapshot_summary(context, project_only, filters)


####################


def snapshot_metadata_get(context, snapshot_id):
    """Get all metadata for a snapshot."""
    return IMPL.snapshot_metadata_get(context, snapshot_id)


def snapshot_metadata_delete(context, snapshot_id, key):
    """Delete the given metadata item."""
    return IMPL.snapshot_metadata_delete(context, snapshot_id, key)


def snapshot_metadata_update(context, snapshot_id, metadata, delete):
    """Update metadata if it exists, otherwise create it."""
    return IMPL.snapshot_metadata_update(context, snapshot_id,
                                         metadata, delete)


####################


def volume_metadata_get(context, volume_id):
    """Get all metadata for a volume."""
    return IMPL.volume_metadata_get(context, volume_id)


def volume_metadata_delete(context, volume_id, key,
                           meta_type=common.METADATA_TYPES.user):
    """Delete the given metadata item."""
    return IMPL.volume_metadata_delete(context, volume_id,
                                       key, meta_type)


def volume_metadata_update(context, volume_id, metadata,
                           delete, meta_type=common.METADATA_TYPES.user):
    """Update metadata if it exists, otherwise create it."""
    return IMPL.volume_metadata_update(context, volume_id, metadata,
                                       delete, meta_type)


##################


def volume_admin_metadata_get(context, volume_id):
    """Get all administration metadata for a volume."""
    return IMPL.volume_admin_metadata_get(context, volume_id)


def volume_admin_metadata_delete(context, volume_id, key):
    """Delete the given metadata item."""
    return IMPL.volume_admin_metadata_delete(context, volume_id, key)


def volume_admin_metadata_update(context, volume_id, metadata, delete,
                                 add=True, update=True):
    """Update metadata if it exists, otherwise create it."""
    return IMPL.volume_admin_metadata_update(context, volume_id, metadata,
                                             delete, add, update)


##################


def volume_type_create(context, values, projects=None):
    """Create a new volume type."""
    return IMPL.volume_type_create(context, values, projects)


def volume_type_update(context, volume_type_id, values):
    return IMPL.volume_type_update(context, volume_type_id, values)


def volume_type_get_all(context, inactive=False, filters=None, marker=None,
                        limit=None, sort_keys=None, sort_dirs=None,
                        offset=None, list_result=False):
    """Get all volume types.

    :param context: context to query under
    :param inactive: Include inactive volume types to the result set
    :param filters: Filters for the query in the form of key/value.
    :param marker: the last item of the previous page, used to determine the
                   next page of results to return
    :param limit: maximum number of items to return
    :param sort_keys: list of attributes by which results should be sorted,
                      paired with corresponding item in sort_dirs
    :param sort_dirs: list of directions in which results should be sorted,
                      paired with corresponding item in sort_keys
    :param list_result: For compatibility, if list_result = True, return a list
                        instead of dict.

        :is_public: Filter volume types based on visibility:

            * **True**: List public volume types only
            * **False**: List private volume types only
            * **None**: List both public and private volume types

    :returns: list/dict of matching volume types
    """

    return IMPL.volume_type_get_all(context, inactive, filters, marker=marker,
                                    limit=limit, sort_keys=sort_keys,
                                    sort_dirs=sort_dirs, offset=offset,
                                    list_result=list_result)


def volume_type_get(context, id, inactive=False, expected_fields=None):
    """Get volume type by id.

    :param context: context to query under
    :param id: Volume type id to get.
    :param inactive: Consider inactive volume types when searching
    :param expected_fields: Return those additional fields.
                            Supported fields are: projects.
    :returns: volume type
    """
    return IMPL.volume_type_get(context, id, inactive, expected_fields)


def volume_type_get_by_name(context, name):
    """Get volume type by name."""
    return IMPL.volume_type_get_by_name(context, name)


def volume_types_get_by_name_or_id(context, volume_type_list):
    """Get volume types by name or id."""
    return IMPL.volume_types_get_by_name_or_id(context, volume_type_list)


def volume_type_qos_associations_get(context, qos_specs_id, inactive=False):
    """Get volume types that are associated with specific qos specs."""
    return IMPL.volume_type_qos_associations_get(context,
                                                 qos_specs_id,
                                                 inactive)


def volume_type_qos_associate(context, type_id, qos_specs_id):
    """Associate a volume type with specific qos specs."""
    return IMPL.volume_type_qos_associate(context, type_id, qos_specs_id)


def volume_type_qos_disassociate(context, qos_specs_id, type_id):
    """Disassociate a volume type from specific qos specs."""
    return IMPL.volume_type_qos_disassociate(context, qos_specs_id, type_id)


def volume_type_qos_disassociate_all(context, qos_specs_id):
    """Disassociate all volume types from specific qos specs."""
    return IMPL.volume_type_qos_disassociate_all(context,
                                                 qos_specs_id)


def volume_type_qos_specs_get(context, type_id):
    """Get all qos specs for given volume type."""
    return IMPL.volume_type_qos_specs_get(context, type_id)


def volume_type_destroy(context, id):
    """Delete a volume type."""
    return IMPL.volume_type_destroy(context, id)


def volume_get_all_active_by_window(context, begin, end=None, project_id=None):
    """Get all the volumes inside the window.

    Specifying a project_id will filter for a certain project.
    """
    return IMPL.volume_get_all_active_by_window(context, begin, end,
                                                project_id)


def volume_type_access_get_all(context, type_id):
    """Get all volume type access of a volume type."""
    return IMPL.volume_type_access_get_all(context, type_id)


def volume_type_access_add(context, type_id, project_id):
    """Add volume type access for project."""
    return IMPL.volume_type_access_add(context, type_id, project_id)


def volume_type_access_remove(context, type_id, project_id):
    """Remove volume type access for project."""
    return IMPL.volume_type_access_remove(context, type_id, project_id)


def project_default_volume_type_set(context, volume_type_id, project_id):
    """Set default volume type for a project"""
    return IMPL.project_default_volume_type_set(context, volume_type_id,
                                                project_id)


def project_default_volume_type_get(context, project_id=None):
    """Get default volume type for a project"""
    return IMPL.project_default_volume_type_get(context, project_id)


def project_default_volume_type_unset(context, project_id):
    """Unset default volume type for a project (hard delete)"""
    return IMPL.project_default_volume_type_unset(context, project_id)


def get_all_projects_with_default_type(context, volume_type_id):
    """Get all the projects associated with a default type"""
    return IMPL.get_all_projects_with_default_type(context, volume_type_id)


####################


def group_type_create(context, values, projects=None):
    """Create a new group type."""
    return IMPL.group_type_create(context, values, projects)


def group_type_update(context, group_type_id, values):
    return IMPL.group_type_update(context, group_type_id, values)


def group_type_get_all(context, inactive=False, filters=None, marker=None,
                       limit=None, sort_keys=None, sort_dirs=None,
                       offset=None, list_result=False):
    """Get all group types.

    :param context: context to query under
    :param inactive: Include inactive group types to the result set
    :param filters: Filters for the query in the form of key/value.
    :param marker: the last item of the previous page, used to determine the
                   next page of results to return
    :param limit: maximum number of items to return
    :param sort_keys: list of attributes by which results should be sorted,
                      paired with corresponding item in sort_dirs
    :param sort_dirs: list of directions in which results should be sorted,
                      paired with corresponding item in sort_keys
    :param list_result: For compatibility, if list_result = True, return a list
                        instead of dict.

        :is_public: Filter group types based on visibility:

            * **True**: List public group types only
            * **False**: List private group types only
            * **None**: List both public and private group types

    :returns: list/dict of matching group types
    """

    return IMPL.group_type_get_all(context, inactive, filters, marker=marker,
                                   limit=limit, sort_keys=sort_keys,
                                   sort_dirs=sort_dirs, offset=offset,
                                   list_result=list_result)


def group_type_get(context, id, inactive=False, expected_fields=None):
    """Get group type by id.

    :param context: context to query under
    :param id: Group type id to get.
    :param inactive: Consider inactive group types when searching
    :param expected_fields: Return those additional fields.
                            Supported fields are: projects.
    :returns: group type
    """
    return IMPL.group_type_get(context, id, inactive, expected_fields)


def group_type_get_by_name(context, name):
    """Get group type by name."""
    return IMPL.group_type_get_by_name(context, name)


def group_types_get_by_name_or_id(context, group_type_list):
    """Get group types by name or id."""
    return IMPL.group_types_get_by_name_or_id(context, group_type_list)


def group_type_destroy(context, id):
    """Delete a group type."""
    return IMPL.group_type_destroy(context, id)


def group_type_access_get_all(context, type_id):
    """Get all group type access of a group type."""
    return IMPL.group_type_access_get_all(context, type_id)


def group_type_access_add(context, type_id, project_id):
    """Add group type access for project."""
    return IMPL.group_type_access_add(context, type_id, project_id)


def group_type_access_remove(context, type_id, project_id):
    """Remove group type access for project."""
    return IMPL.group_type_access_remove(context, type_id, project_id)


def volume_type_get_all_by_group(context, group_id):
    """Get all volumes in a group."""
    return IMPL.volume_type_get_all_by_group(context, group_id)


####################


def volume_type_extra_specs_get(context, volume_type_id):
    """Get all extra specs for a volume type."""
    return IMPL.volume_type_extra_specs_get(context, volume_type_id)


def volume_type_extra_specs_delete(context, volume_type_id, key):
    """Delete the given extra specs item."""
    return IMPL.volume_type_extra_specs_delete(context, volume_type_id, key)


def volume_type_extra_specs_update_or_create(context,
                                             volume_type_id,
                                             extra_specs):
    """Create or update volume type extra specs.

    This adds or modifies the key/value pairs specified in the extra specs dict
    argument.
    """
    return IMPL.volume_type_extra_specs_update_or_create(context,
                                                         volume_type_id,
                                                         extra_specs)


###################


def group_type_specs_get(context, group_type_id):
    """Get all group specs for a group type."""
    return IMPL.group_type_specs_get(context, group_type_id)


def group_type_specs_delete(context, group_type_id, key):
    """Delete the given group specs item."""
    return IMPL.group_type_specs_delete(context, group_type_id, key)


def group_type_specs_update_or_create(context,
                                      group_type_id,
                                      group_specs):
    """Create or update group type specs.

    This adds or modifies the key/value pairs specified in the group specs dict
    argument.
    """
    return IMPL.group_type_specs_update_or_create(context,
                                                  group_type_id,
                                                  group_specs)


###################


def volume_type_encryption_get(context, volume_type_id, session=None):
    return IMPL.volume_type_encryption_get(context, volume_type_id, session)


def volume_type_encryption_delete(context, volume_type_id):
    return IMPL.volume_type_encryption_delete(context, volume_type_id)


def volume_type_encryption_create(context, volume_type_id, encryption_specs):
    return IMPL.volume_type_encryption_create(context, volume_type_id,
                                              encryption_specs)


def volume_type_encryption_update(context, volume_type_id, encryption_specs):
    return IMPL.volume_type_encryption_update(context, volume_type_id,
                                              encryption_specs)


def volume_type_encryption_volume_get(context, volume_type_id, session=None):
    return IMPL.volume_type_encryption_volume_get(context, volume_type_id,
                                                  session)


def volume_encryption_metadata_get(context, volume_id, session=None):
    return IMPL.volume_encryption_metadata_get(context, volume_id, session)


###################


def qos_specs_create(context, values):
    """Create a qos_specs."""
    return IMPL.qos_specs_create(context, values)


def qos_specs_get(context, qos_specs_id):
    """Get all specification for a given qos_specs."""
    return IMPL.qos_specs_get(context, qos_specs_id)


def qos_specs_get_all(context, filters=None, marker=None, limit=None,
                      offset=None, sort_keys=None, sort_dirs=None):
    """Get all qos_specs."""
    return IMPL.qos_specs_get_all(context, filters=filters, marker=marker,
                                  limit=limit, offset=offset,
                                  sort_keys=sort_keys, sort_dirs=sort_dirs)


def qos_specs_get_by_name(context, name):
    """Get all specification for a given qos_specs."""
    return IMPL.qos_specs_get_by_name(context, name)


def qos_specs_associations_get(context, qos_specs_id):
    """Get all associated volume types for a given qos_specs."""
    return IMPL.qos_specs_associations_get(context, qos_specs_id)


def qos_specs_associate(context, qos_specs_id, type_id):
    """Associate qos_specs from volume type."""
    return IMPL.qos_specs_associate(context, qos_specs_id, type_id)


def qos_specs_disassociate(context, qos_specs_id, type_id):
    """Disassociate qos_specs from volume type."""
    return IMPL.qos_specs_disassociate(context, qos_specs_id, type_id)


def qos_specs_disassociate_all(context, qos_specs_id):
    """Disassociate qos_specs from all entities."""
    return IMPL.qos_specs_disassociate_all(context, qos_specs_id)


def qos_specs_delete(context, qos_specs_id):
    """Delete the qos_specs."""
    return IMPL.qos_specs_delete(context, qos_specs_id)


def qos_specs_item_delete(context, qos_specs_id, key):
    """Delete specified key in the qos_specs."""
    return IMPL.qos_specs_item_delete(context, qos_specs_id, key)


def qos_specs_update(context, qos_specs_id, specs):
    """Update qos specs.

    This adds or modifies the key/value pairs specified in the
    specs dict argument for a given qos_specs.
    """
    return IMPL.qos_specs_update(context, qos_specs_id, specs)


###################


def volume_glance_metadata_create(context, volume_id, key, value):
    """Update the Glance metadata for the specified volume."""
    return IMPL.volume_glance_metadata_create(context,
                                              volume_id,
                                              key,
                                              value)


def volume_glance_metadata_bulk_create(context, volume_id, metadata):
    """Add Glance metadata for specified volume (multiple pairs)."""
    return IMPL.volume_glance_metadata_bulk_create(context, volume_id,
                                                   metadata)


def volume_glance_metadata_get_all(context):
    """Return the glance metadata for all volumes."""
    return IMPL.volume_glance_metadata_get_all(context)


def volume_glance_metadata_get(context, volume_id):
    """Return the glance metadata for a volume."""
    return IMPL.volume_glance_metadata_get(context, volume_id)


def volume_glance_metadata_list_get(context, volume_id_list):
    """Return the glance metadata for a volume list."""
    return IMPL.volume_glance_metadata_list_get(context, volume_id_list)


def volume_snapshot_glance_metadata_get(context, snapshot_id):
    """Return the Glance metadata for the specified snapshot."""
    return IMPL.volume_snapshot_glance_metadata_get(context, snapshot_id)


def volume_glance_metadata_copy_to_snapshot(context, snapshot_id, volume_id):
    """Update the Glance metadata for a snapshot.

    This will copy all of the key:value pairs from the originating volume,
    to ensure that a volume created from the snapshot will retain the
    original metadata.
    """
    return IMPL.volume_glance_metadata_copy_to_snapshot(context, snapshot_id,
                                                        volume_id)


def volume_glance_metadata_copy_to_volume(context, volume_id, snapshot_id):
    """Update the Glance metadata from a volume (created from a snapshot).

    This will copy all of the key:value pairs from the originating snapshot,
    to ensure that the Glance metadata from the original volume is retained.
    """
    return IMPL.volume_glance_metadata_copy_to_volume(context, volume_id,
                                                      snapshot_id)


def volume_glance_metadata_delete_by_volume(context, volume_id):
    """Delete the glance metadata for a volume."""
    return IMPL.volume_glance_metadata_delete_by_volume(context, volume_id)


def volume_glance_metadata_delete_by_snapshot(context, snapshot_id):
    """Delete the glance metadata for a snapshot."""
    return IMPL.volume_glance_metadata_delete_by_snapshot(context, snapshot_id)


def volume_glance_metadata_copy_from_volume_to_volume(context,
                                                      src_volume_id,
                                                      volume_id):
    """Update the Glance metadata for a volume.

    Update the Glance metadata for a volume by copying all of the key:value
    pairs from the originating volume.

    This is so that a volume created from the volume (clone) will retain the
    original metadata.
    """
    return IMPL.volume_glance_metadata_copy_from_volume_to_volume(
        context,
        src_volume_id,
        volume_id)


###################


def quota_create(context, project_id, resource, limit):
    """Create a quota for the given project and resource."""
    return IMPL.quota_create(context, project_id, resource, limit)


def quota_get(context, project_id, resource):
    """Retrieve a quota or raise if it does not exist."""
    return IMPL.quota_get(context, project_id, resource)


def quota_get_all_by_project(context, project_id):
    """Retrieve all quotas associated with a given project."""
    return IMPL.quota_get_all_by_project(context, project_id)


def quota_update(context, project_id, resource, limit):
    """Update a quota or raise if it does not exist."""
    return IMPL.quota_update(context, project_id, resource, limit)


def quota_update_resource(context, old_res, new_res):
    """Update resource of quotas."""
    return IMPL.quota_update_resource(context, old_res, new_res)


def quota_destroy(context, project_id, resource):
    """Destroy the quota or raise if it does not exist."""
    return IMPL.quota_destroy(context, project_id, resource)


###################


def quota_class_create(context, class_name, resource, limit):
    """Create a quota class for the given name and resource."""
    return IMPL.quota_class_create(context, class_name, resource, limit)


def quota_class_get(context, class_name, resource):
    """Retrieve a quota class or raise if it does not exist."""
    return IMPL.quota_class_get(context, class_name, resource)


def quota_class_get_defaults(context):
    """Retrieve all default quotas."""
    return IMPL.quota_class_get_defaults(context)


def quota_class_get_all_by_name(context, class_name):
    """Retrieve all quotas associated with a given quota class."""
    return IMPL.quota_class_get_all_by_name(context, class_name)


def quota_class_update(context, class_name, resource, limit):
    """Update a quota class or raise if it does not exist."""
    return IMPL.quota_class_update(context, class_name, resource, limit)


def quota_class_update_resource(context, resource, new_resource):
    """Update resource name in quota_class."""
    return IMPL.quota_class_update_resource(context, resource, new_resource)


def quota_class_destroy(context, class_name, resource):
    """Destroy the quota class or raise if it does not exist."""
    return IMPL.quota_class_destroy(context, class_name, resource)


def quota_class_destroy_all_by_name(context, class_name):
    """Destroy all quotas associated with a given quota class."""
    return IMPL.quota_class_destroy_all_by_name(context, class_name)


###################


def quota_usage_get(context, project_id, resource):
    """Retrieve a quota usage or raise if it does not exist."""
    return IMPL.quota_usage_get(context, project_id, resource)


def quota_usage_get_all_by_project(context, project_id):
    """Retrieve all usage associated with a given resource."""
    return IMPL.quota_usage_get_all_by_project(context, project_id)


###################


def quota_reserve(context, resources, quotas, deltas, expire,
                  until_refresh, max_age, project_id=None):
    """Check quotas and create appropriate reservations."""
    return IMPL.quota_reserve(context, resources, quotas, deltas, expire,
                              until_refresh, max_age, project_id=project_id)


def reservation_commit(context, reservations, project_id=None):
    """Commit quota reservations."""
    return IMPL.reservation_commit(context, reservations,
                                   project_id=project_id)


def reservation_rollback(context, reservations, project_id=None):
    """Roll back quota reservations."""
    return IMPL.reservation_rollback(context, reservations,
                                     project_id=project_id)


def quota_destroy_by_project(context, project_id):
    """Destroy all quotas associated with a given project."""
    return IMPL.quota_destroy_by_project(context, project_id)


def reservation_expire(context):
    """Roll back any expired reservations."""
    return IMPL.reservation_expire(context)


def quota_usage_update_resource(context, old_res, new_res):
    """Update resource field in quota_usages."""
    return IMPL.quota_usage_update_resource(context, old_res, new_res)


###################


def backup_get(context, backup_id, read_deleted=None, project_only=True):
    """Get a backup or raise if it does not exist."""
    return IMPL.backup_get(context, backup_id, read_deleted, project_only)


def backup_get_all(context, filters=None, marker=None, limit=None,
                   offset=None, sort_keys=None, sort_dirs=None):
    """Get all backups."""
    return IMPL.backup_get_all(context, filters=filters, marker=marker,
                               limit=limit, offset=offset, sort_keys=sort_keys,
                               sort_dirs=sort_dirs)


def backup_get_all_by_host(context, host):
    """Get all backups belonging to a host."""
    return IMPL.backup_get_all_by_host(context, host)


def backup_create(context, values):
    """Create a backup from the values dictionary."""
    return IMPL.backup_create(context, values)


def backup_metadata_get(context, backup_id):
    return IMPL.backup_metadata_get(context, backup_id)


def backup_metadata_update(context, backup_id, metadata, delete):
    return IMPL.backup_metadata_update(context, backup_id, metadata, delete)


def backup_get_all_by_project(context, project_id, filters=None, marker=None,
                              limit=None, offset=None, sort_keys=None,
                              sort_dirs=None):
    """Get all backups belonging to a project."""
    return IMPL.backup_get_all_by_project(context, project_id,
                                          filters=filters, marker=marker,
                                          limit=limit, offset=offset,
                                          sort_keys=sort_keys,
                                          sort_dirs=sort_dirs)


def backup_get_all_by_volume(context, volume_id, vol_project_id, filters=None):
    """Get all backups belonging to a volume."""
    return IMPL.backup_get_all_by_volume(context, volume_id, vol_project_id,
                                         filters=filters)


def backup_get_all_active_by_window(context, begin, end=None, project_id=None):
    """Get all the backups inside the window.

    Specifying a project_id will filter for a certain project.
    """
    return IMPL.backup_get_all_active_by_window(context, begin, end,
                                                project_id)


def backup_update(context, backup_id, values):
    """Set the given properties on a backup and update it.

    Raises NotFound if backup does not exist.
    """
    return IMPL.backup_update(context, backup_id, values)


def backup_destroy(context, backup_id):
    """Destroy the backup or raise if it does not exist."""
    return IMPL.backup_destroy(context, backup_id)


###################


def transfer_get(context, transfer_id):
    """Get a volume transfer record or raise if it does not exist."""
    return IMPL.transfer_get(context, transfer_id)


def transfer_get_all(context, marker=None, limit=None, sort_keys=None,
                     sort_dirs=None, filters=None, offset=None):
    """Get all volume transfer records."""
    return IMPL.transfer_get_all(context, marker=marker, limit=limit,
                                 sort_keys=sort_keys, sort_dirs=sort_dirs,
                                 filters=filters, offset=offset)


def transfer_get_all_by_project(context, project_id, marker=None,
                                limit=None, sort_keys=None,
                                sort_dirs=None, filters=None, offset=None):
    """Get all volume transfer records for specified project."""
    return IMPL.transfer_get_all_by_project(context, project_id, marker=marker,
                                            limit=limit, sort_keys=sort_keys,
                                            sort_dirs=sort_dirs,
                                            filters=filters, offset=offset)


def transfer_create(context, values):
    """Create an entry in the transfers table."""
    return IMPL.transfer_create(context, values)


def transfer_destroy(context, transfer_id):
    """Destroy a record in the volume transfer table."""
    return IMPL.transfer_destroy(context, transfer_id)


def transfer_accept(context, transfer_id, user_id, project_id,
                    no_snapshots=False):
    """Accept a volume transfer."""
    return IMPL.transfer_accept(context, transfer_id, user_id, project_id,
                                no_snapshots=no_snapshots)


###################


def consistencygroup_get(context, consistencygroup_id):
    """Get a consistencygroup or raise if it does not exist."""
    return IMPL.consistencygroup_get(context, consistencygroup_id)


def consistencygroup_get_all(context, filters=None, marker=None, limit=None,
                             offset=None, sort_keys=None, sort_dirs=None):
    """Get all consistencygroups."""
    return IMPL.consistencygroup_get_all(context, filters=filters,
                                         marker=marker, limit=limit,
                                         offset=offset, sort_keys=sort_keys,
                                         sort_dirs=sort_dirs)


def consistencygroup_create(context, values, cg_snap_id=None, cg_id=None):
    """Create a consistencygroup from the values dictionary."""
    return IMPL.consistencygroup_create(context, values, cg_snap_id, cg_id)


def consistencygroup_get_all_by_project(context, project_id, filters=None,
                                        marker=None, limit=None, offset=None,
                                        sort_keys=None, sort_dirs=None):
    """Get all consistencygroups belonging to a project."""
    return IMPL.consistencygroup_get_all_by_project(context, project_id,
                                                    filters=filters,
                                                    marker=marker, limit=limit,
                                                    offset=offset,
                                                    sort_keys=sort_keys,
                                                    sort_dirs=sort_dirs)


def consistencygroup_update(context, consistencygroup_id, values):
    """Set the given properties on a consistencygroup and update it.

    Raises NotFound if consistencygroup does not exist.
    """
    return IMPL.consistencygroup_update(context, consistencygroup_id, values)


def consistencygroup_destroy(context, consistencygroup_id):
    """Destroy the consistencygroup or raise if it does not exist."""
    return IMPL.consistencygroup_destroy(context, consistencygroup_id)


def cg_has_cgsnapshot_filter():
    """Return a filter that checks if a CG has CG Snapshots."""
    return IMPL.cg_has_cgsnapshot_filter()


def cg_has_volumes_filter(attached_or_with_snapshots=False):
    """Return a filter to check if a CG has volumes.

    When attached_or_with_snapshots parameter is given a True value only
    attached volumes or those with snapshots will be considered.
    """
    return IMPL.cg_has_volumes_filter(attached_or_with_snapshots)


def cg_creating_from_src(cg_id=None, cgsnapshot_id=None):
    """Return a filter to check if a CG is being used as creation source.

    Returned filter is meant to be used in the Conditional Update mechanism and
    checks if provided CG ID or CG Snapshot ID is currently being used to
    create another CG.

    This filter will not include CGs that have used the ID but have already
    finished their creation (status is no longer creating).

    Filter uses a subquery that allows it to be used on updates to the
    consistencygroups table.
    """
    return IMPL.cg_creating_from_src(cg_id, cgsnapshot_id)


def consistencygroup_include_in_cluster(context, cluster, partial_rename=True,
                                        **filters):
    """Include all consistency groups matching the filters into a cluster.

    When partial_rename is set we will not set the cluster_name with cluster
    parameter value directly, we'll replace provided cluster_name or host
    filter value with cluster instead.

    This is useful when we want to replace just the cluster name but leave
    the backend and pool information as it is.  If we are using cluster_name
    to filter, we'll use that same DB field to replace the cluster value and
    leave the rest as it is.  Likewise if we use the host to filter.

    Returns the number of consistency groups that have been changed.
    """
    return IMPL.consistencygroup_include_in_cluster(context, cluster,
                                                    partial_rename,
                                                    **filters)


def group_include_in_cluster(context, cluster, partial_rename=True, **filters):
    """Include all generic groups matching the filters into a cluster.

    When partial_rename is set we will not set the cluster_name with cluster
    parameter value directly, we'll replace provided cluster_name or host
    filter value with cluster instead.

    This is useful when we want to replace just the cluster name but leave
    the backend and pool information as it is.  If we are using cluster_name
    to filter, we'll use that same DB field to replace the cluster value and
    leave the rest as it is.  Likewise if we use the host to filter.

    Returns the number of generic groups that have been changed.
    """
    return IMPL.group_include_in_cluster(context, cluster, partial_rename,
                                         **filters)

###################


def group_get(context, group_id):
    """Get a group or raise if it does not exist."""
    return IMPL.group_get(context, group_id)


def group_get_all(context, filters=None, marker=None, limit=None,
                  offset=None, sort_keys=None, sort_dirs=None):
    """Get all groups."""
    return IMPL.group_get_all(context, filters=filters,
                              marker=marker, limit=limit,
                              offset=offset, sort_keys=sort_keys,
                              sort_dirs=sort_dirs)


def group_create(context, values, group_snapshot_id=None, group_id=None):
    """Create a group from the values dictionary."""
    return IMPL.group_create(context, values, group_snapshot_id, group_id)


def group_get_all_by_project(context, project_id, filters=None,
                             marker=None, limit=None, offset=None,
                             sort_keys=None, sort_dirs=None):
    """Get all groups belonging to a project."""
    return IMPL.group_get_all_by_project(context, project_id,
                                         filters=filters,
                                         marker=marker, limit=limit,
                                         offset=offset,
                                         sort_keys=sort_keys,
                                         sort_dirs=sort_dirs)


def group_update(context, group_id, values):
    """Set the given properties on a group and update it.

    Raises NotFound if group does not exist.
    """
    return IMPL.group_update(context, group_id, values)


def group_destroy(context, group_id):
    """Destroy the group or raise if it does not exist."""
    return IMPL.group_destroy(context, group_id)


def group_has_group_snapshot_filter():
    """Return a filter that checks if a Group has Group Snapshots."""
    return IMPL.group_has_group_snapshot_filter()


def group_has_volumes_filter(attached_or_with_snapshots=False):
    """Return a filter to check if a Group has volumes.

    When attached_or_with_snapshots parameter is given a True value only
    attached volumes or those with snapshots will be considered.
    """
    return IMPL.group_has_volumes_filter(attached_or_with_snapshots)


def group_creating_from_src(group_id=None, group_snapshot_id=None):
    """Return a filter to check if a Group is being used as creation source.

    Returned filter is meant to be used in the Conditional Update mechanism and
    checks if provided Group ID or Group Snapshot ID is currently being used to
    create another Group.

    This filter will not include Groups that have used the ID but have already
    finished their creation (status is no longer creating).

    Filter uses a subquery that allows it to be used on updates to the
    groups table.
    """
    return IMPL.group_creating_from_src(group_id, group_snapshot_id)


def group_volume_type_mapping_create(context, group_id, volume_type_id):
    """Create a group volume_type mapping entry."""
    return IMPL.group_volume_type_mapping_create(context, group_id,
                                                 volume_type_id)


###################


def cgsnapshot_get(context, cgsnapshot_id):
    """Get a cgsnapshot or raise if it does not exist."""
    return IMPL.cgsnapshot_get(context, cgsnapshot_id)


def cgsnapshot_get_all(context, filters=None):
    """Get all cgsnapshots."""
    return IMPL.cgsnapshot_get_all(context, filters)


def cgsnapshot_create(context, values):
    """Create a cgsnapshot from the values dictionary."""
    return IMPL.cgsnapshot_create(context, values)


def cgsnapshot_get_all_by_group(context, group_id, filters=None):
    """Get all cgsnapshots belonging to a consistency group."""
    return IMPL.cgsnapshot_get_all_by_group(context, group_id, filters)


def cgsnapshot_get_all_by_project(context, project_id, filters=None):
    """Get all cgsnapshots belonging to a project."""
    return IMPL.cgsnapshot_get_all_by_project(context, project_id, filters)


def cgsnapshot_update(context, cgsnapshot_id, values):
    """Set the given properties on a cgsnapshot and update it.

    Raises NotFound if cgsnapshot does not exist.
    """
    return IMPL.cgsnapshot_update(context, cgsnapshot_id, values)


def cgsnapshot_destroy(context, cgsnapshot_id):
    """Destroy the cgsnapshot or raise if it does not exist."""
    return IMPL.cgsnapshot_destroy(context, cgsnapshot_id)


def cgsnapshot_creating_from_src():
    """Get a filter that checks if a CGSnapshot is being created from a CG."""
    return IMPL.cgsnapshot_creating_from_src()


###################


def group_snapshot_get(context, group_snapshot_id):
    """Get a group snapshot or raise if it does not exist."""
    return IMPL.group_snapshot_get(context, group_snapshot_id)


def group_snapshot_get_all(context, filters=None, marker=None, limit=None,
                           offset=None, sort_keys=None, sort_dirs=None):
    """Get all group snapshots."""
    return IMPL.group_snapshot_get_all(context, filters, marker, limit,
                                       offset, sort_keys, sort_dirs)


def group_snapshot_create(context, values):
    """Create a group snapshot from the values dictionary."""
    return IMPL.group_snapshot_create(context, values)


def group_snapshot_get_all_by_group(context, group_id, filters=None,
                                    marker=None, limit=None,
                                    offset=None, sort_keys=None,
                                    sort_dirs=None):
    """Get all group snapshots belonging to a group."""
    return IMPL.group_snapshot_get_all_by_group(context, group_id,
                                                filters, marker, limit,
                                                offset, sort_keys, sort_dirs)


def group_snapshot_get_all_by_project(context, project_id, filters=None,
                                      marker=None, limit=None,
                                      offset=None, sort_keys=None,
                                      sort_dirs=None):
    """Get all group snapshots belonging to a project."""
    return IMPL.group_snapshot_get_all_by_project(context, project_id,
                                                  filters, marker, limit,
                                                  offset, sort_keys, sort_dirs)


def group_snapshot_update(context, group_snapshot_id, values):
    """Set the given properties on a group snapshot and update it.

    Raises NotFound if group snapshot does not exist.
    """
    return IMPL.group_snapshot_update(context, group_snapshot_id, values)


def group_snapshot_destroy(context, group_snapshot_id):
    """Destroy the group snapshot or raise if it does not exist."""
    return IMPL.group_snapshot_destroy(context, group_snapshot_id)


def group_snapshot_creating_from_src():
    """Get a filter to check if a grp snapshot is being created from a grp."""
    return IMPL.group_snapshot_creating_from_src()


###################


def purge_deleted_rows(context, age_in_days):
    """Purge deleted rows older than given age from cinder tables

    Raises InvalidParameterValue if age_in_days is incorrect.
    :returns: number of deleted rows
    """
    return IMPL.purge_deleted_rows(context, age_in_days=age_in_days)


def get_booleans_for_table(table_name):
    return IMPL.get_booleans_for_table(table_name)


###################


def reset_active_backend(context, enable_replication, active_backend_id,
                         backend_host):
    """Reset the active backend for a host."""
    return IMPL.reset_active_backend(context, enable_replication,
                                     active_backend_id, backend_host)


###################


def driver_initiator_data_insert_by_key(context, initiator,
                                        namespace, key, value):
    """Updates DriverInitiatorData entry.

    Sets the value for the specified key within the namespace.

    If the entry already exists return False, if it inserted successfully
    return True.
    """
    return IMPL.driver_initiator_data_insert_by_key(context,
                                                    initiator,
                                                    namespace,
                                                    key,
                                                    value)


def driver_initiator_data_get(context, initiator, namespace):
    """Query for an DriverInitiatorData that has the specified key"""
    return IMPL.driver_initiator_data_get(context, initiator, namespace)


###################


def image_volume_cache_create(context, host, cluster_name, image_id,
                              image_updated_at, volume_id, size):
    """Create a new image volume cache entry."""
    return IMPL.image_volume_cache_create(context,
                                          host,
                                          cluster_name,
                                          image_id,
                                          image_updated_at,
                                          volume_id,
                                          size)


def image_volume_cache_delete(context, volume_id):
    """Delete an image volume cache entry specified by volume id."""
    return IMPL.image_volume_cache_delete(context, volume_id)


def image_volume_cache_get_and_update_last_used(context, image_id, **filters):
    """Query for an image volume cache entry."""
    return IMPL.image_volume_cache_get_and_update_last_used(context,
                                                            image_id,
                                                            **filters)


def image_volume_cache_get_by_volume_id(context, volume_id):
    """Query to see if a volume id is an image-volume contained in the cache"""
    return IMPL.image_volume_cache_get_by_volume_id(context, volume_id)


def image_volume_cache_get_all(context, **filters):
    """Query for all image volume cache entry for a host."""
    return IMPL.image_volume_cache_get_all(context, **filters)


def image_volume_cache_include_in_cluster(context, cluster,
                                          partial_rename=True, **filters):
    """Include in cluster image volume cache entries matching the filters.

    When partial_rename is set we will not set the cluster_name with cluster
    parameter value directly, we'll replace provided cluster_name or host
    filter value with cluster instead.

    This is useful when we want to replace just the cluster name but leave
    the backend and pool information as it is.  If we are using cluster_name
    to filter, we'll use that same DB field to replace the cluster value and
    leave the rest as it is.  Likewise if we use the host to filter.

    Returns the number of volumes that have been changed.
    """
    return IMPL.image_volume_cache_include_in_cluster(
        context, cluster, partial_rename, **filters)


###################


def message_get(context, message_id):
    """Return a message with the specified ID."""
    return IMPL.message_get(context, message_id)


def message_get_all(context, filters=None, marker=None, limit=None,
                    offset=None, sort_keys=None, sort_dirs=None):
    return IMPL.message_get_all(context, filters=filters, marker=marker,
                                limit=limit, offset=offset,
                                sort_keys=sort_keys, sort_dirs=sort_dirs)


def message_create(context, values):
    """Creates a new message with the specified values."""
    return IMPL.message_create(context, values)


def message_destroy(context, message_id):
    """Deletes message with the specified ID."""
    return IMPL.message_destroy(context, message_id)


def cleanup_expired_messages(context):
    """Soft delete expired messages"""
    return IMPL.cleanup_expired_messages(context)


###################


def workers_init():
    """Check if DB supports subsecond resolution and set global flag.

    MySQL 5.5 doesn't support subsecond resolution in datetime fields, so we
    have to take it into account when working with the worker's table.

    Once we drop support for MySQL 5.5 we can remove this method.
    """
    return IMPL.workers_init()


def worker_create(context, **values):
    """Create a worker entry from optional arguments."""
    return IMPL.worker_create(context, **values)


def worker_get(context, **filters):
    """Get a worker or raise exception if it does not exist."""
    return IMPL.worker_get(context, **filters)


def worker_get_all(context, until=None, db_filters=None, **filters):
    """Get all workers that match given criteria."""
    return IMPL.worker_get_all(context, until=until, db_filters=db_filters,
                               **filters)


def worker_update(context, id, filters=None, orm_worker=None, **values):
    """Update a worker with given values."""
    return IMPL.worker_update(context, id, filters=filters,
                              orm_worker=orm_worker, **values)


def worker_claim_for_cleanup(context, claimer_id, orm_worker):
    """Soft delete a worker, change the service_id and update the worker."""
    return IMPL.worker_claim_for_cleanup(context, claimer_id, orm_worker)


def worker_destroy(context, **filters):
    """Delete a worker (no soft delete)."""
    return IMPL.worker_destroy(context, **filters)


###################


def resource_exists(context, model, resource_id):
    return IMPL.resource_exists(context, model, resource_id)


def get_model_for_versioned_object(versioned_object):
    return IMPL.get_model_for_versioned_object(versioned_object)


def get_by_id(context, model, id, *args, **kwargs):
    return IMPL.get_by_id(context, model, id, *args, **kwargs)


class Condition(object):
    """Class for normal condition values for conditional_update."""
    def __init__(self, value, field=None):
        self.value = value
        # Field is optional and can be passed when getting the filter
        self.field = field

    def get_filter(self, model, field=None):
        return IMPL.condition_db_filter(model, self._get_field(field),
                                        self.value)

    def _get_field(self, field=None):
        # We must have a defined field on initialization or when called
        field = field or self.field
        if not field:
            raise ValueError(_('Condition has no field.'))
        return field

###################


def attachment_specs_exist(context):
    """Check if there are attachment specs left."""
    return IMPL.attachment_specs_exist(context)


def attachment_specs_get(context, attachment_id):
    """DEPRECATED: Get all specs for an attachment."""
    return IMPL.attachment_specs_get(context, attachment_id)


def attachment_specs_delete(context, attachment_id, key):
    """DEPRECATED: Delete the given attachment specs item."""
    return IMPL.attachment_specs_delete(context, attachment_id, key)


def attachment_specs_update_or_create(context,
                                      attachment_id,
                                      specs):
    """DEPRECATED: Create or update attachment specs.

    This adds or modifies the key/value pairs specified in the attachment
    specs dict argument.
    """
    return IMPL.attachment_specs_update_or_create(context,
                                                  attachment_id,
                                                  specs)


###################


class Not(Condition):
    """Class for negated condition values for conditional_update.

    By default NULL values will be treated like Python treats None instead of
    how SQL treats it.

    So for example when values are (1, 2) it will evaluate to True when we have
    value 3 or NULL, instead of only with 3 like SQL does.
    """
    def __init__(self, value, field=None, auto_none=True):
        super(Not, self).__init__(value, field)
        self.auto_none = auto_none

    def get_filter(self, model, field=None):
        # If implementation has a specific method use it
        if hasattr(IMPL, 'condition_not_db_filter'):
            return IMPL.condition_not_db_filter(model, self._get_field(field),
                                                self.value, self.auto_none)

        # Otherwise non negated object must adming ~ operator for not
        return ~super(Not, self).get_filter(model, field)


class Case(object):
    """Class for conditional value selection for conditional_update."""
    def __init__(self, whens, value=None, else_=None):
        self.whens = whens
        self.value = value
        self.else_ = else_


def is_orm_value(obj):
    """Check if object is an ORM field."""
    return IMPL.is_orm_value(obj)


def conditional_update(context, model, values, expected_values, filters=(),
                       include_deleted='no', project_only=False, order=None):
    """Compare-and-swap conditional update.

    Update will only occur in the DB if conditions are met.

    We have 4 different condition types we can use in expected_values:
     - Equality:  {'status': 'available'}
     - Inequality: {'status': vol_obj.Not('deleting')}
     - In range: {'status': ['available', 'error']
     - Not in range: {'status': vol_obj.Not(['in-use', 'attaching'])

    Method accepts additional filters, which are basically anything that can be
    passed to a sqlalchemy query's filter method, for example:

    .. code-block:: python

     [~sql.exists().where(models.Volume.id == models.Snapshot.volume_id)]

    We can select values based on conditions using Case objects in the 'values'
    argument. For example:

    .. code-block:: python

     has_snapshot_filter = sql.exists().where(
         models.Snapshot.volume_id == models.Volume.id)
     case_values = db.Case([(has_snapshot_filter, 'has-snapshot')],
                           else_='no-snapshot')
     db.conditional_update(context, models.Volume, {'status': case_values},
                           {'status': 'available'})

    And we can use DB fields for example to store previous status in the
    corresponding field even though we don't know which value is in the db from
    those we allowed:

    .. code-block:: python

     db.conditional_update(context, models.Volume,
                           {'status': 'deleting',
                            'previous_status': models.Volume.status},
                           {'status': ('available', 'error')})

    :param values: Dictionary of key-values to update in the DB.
    :param expected_values: Dictionary of conditions that must be met for the
                            update to be executed.
    :param filters: Iterable with additional filters.
    :param include_deleted: Should the update include deleted items, this is
                            equivalent to read_deleted.
    :param project_only: Should the query be limited to context's project.
    :param order: Specific order of fields in which to update the values
    :returns: Number of db rows that were updated.
    """
    return IMPL.conditional_update(context, model, values, expected_values,
                                   filters, include_deleted, project_only,
                                   order)


# TODO: (Y Release) remove method and this comment
def volume_use_quota_online_data_migration(context, max_count):
    return IMPL.volume_use_quota_online_data_migration(context, max_count)


# TODO: (Y Release) remove method and this comment
def snapshot_use_quota_online_data_migration(context, max_count):
    return IMPL.snapshot_use_quota_online_data_migration(context, max_count)


# TODO: (Z Release) remove method and this comment
# TODO: (Y Release) uncomment method
# def remove_temporary_admin_metadata_data_migration(context, max_count):
#     IMPL.remove_temporary_admin_metadata_data_migration(context, max_count)
